# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TAO dataset."""

from __future__ import annotations

import collections
import json
import os
from typing import Any, Dict, Optional, Tuple

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_VIDEO_URL = 'https://motchallenge.net/data/'
_ANNOTATIONS_URL = (
    'https://github.com/TAO-Dataset/annotations/archive/v1.2.tar.gz'
)

_DESCRIPTION = """
The TAO dataset is a large video object detection dataset consisting of
2,907 high resolution videos and 833 object categories. Note that this dataset
requires at least 300 GB of free space to store.
"""

_CITATION = """
@article{Dave_2020,
   title={TAO: A Large-Scale Benchmark for Tracking Any Object},
   ISBN={9783030585587},
   ISSN={1611-3349},
   url={http://dx.doi.org/10.1007/978-3-030-58558-7_26},
   DOI={10.1007/978-3-030-58558-7_26},
   journal={Lecture Notes in Computer Science},
   publisher={Springer International Publishing},
   author={Dave, Achal and Khurana, Tarasha and Tokmakov, Pavel and Schmid, Cordelia and Ramanan, Deva},
   year={2020},
   pages={436-454}
}
"""

NestedDict = Dict[str, Any]


def _build_annotations_index(
    annotations: NestedDict,
) -> Tuple[NestedDict, NestedDict, NestedDict, NestedDict]:
  """Builds several dictionaries to aid in looking up annotations."""
  vids = {x['id']: x for x in annotations['videos']}
  images = {x['id']: x for x in annotations['images']}
  ann_to_images = {}
  track_to_anns = collections.defaultdict(list)
  vid_to_tracks = collections.defaultdict(list)
  for track in annotations['tracks']:
    vid_to_tracks[track['video_id']].append(track)
  for ann in annotations['annotations']:
    ann_to_images[ann['id']] = images[ann['image_id']]
    track_to_anns[ann['track_id']].append(ann)
  return vids, ann_to_images, track_to_anns, vid_to_tracks


def _merge_categories_map(annotations: NestedDict) -> Dict[str, str]:
  """Some categories should be renamed into others.

  This code segment is based on the TAO provided preprocessing API.

  Args:
    annotations: a dictionary containing all the annotations

  Returns:
    merge_map: dictionary mapping from category id to merged id
  """
  merge_map = {}
  for category in annotations['categories']:
    if 'merged' in category:
      for to_merge in category['merged']:
        merge_map[to_merge['id']] = category['id']
  return merge_map


def _maybe_prepare_manual_data(dl_manager: tfds.download.DownloadManager):
  """Return paths to the manually downloaded data if it is available."""

  # The file has a different name each time it is downloaded.
  manually_downloaded_files = [
      '1_AVA_HACS_TRAIN_*.zip',
      '2_AVA_HACS_VAL_*.zip',
  ]
  files = []
  for file in manually_downloaded_files:
    file_glob = [_ for _ in dl_manager.manual_dir.glob(file)]
    if not file_glob:  # No manually downloaded files.
      return None, None
    if len(file_glob) == 1:
      files.append(file_glob[0])
    else:
      raise ValueError(
          'Unexpected multiple files matching pattern: '
          f'{file} inside {os.fspath(dl_manager.manual_dir)}. '
          'There should only be one file matching this pattern.'
      )
  return dl_manager.extract(files)


def _get_category_id_map(annotations_root) -> Dict[str, int]:
  """Gets a map from the TAO category id to a tfds category index.

  The tfds category index is the index which a category appears in the
  label list.

  Args:
    annotations_root: directory containing the train and validation annotations.

  Returns:
    id_map: A dict mapping from TAO category id to tfds category index,
      filtered to contain only categories appearing in the train and val set.
  """

  train = json.loads((annotations_root / 'train.json').read_text())
  val = json.loads((annotations_root / 'validation.json').read_text())

  merge_map = _merge_categories_map(train)
  merge_map.update(_merge_categories_map(val))

  classes = set()
  classes |= {a['category_id'] for a in train['annotations']}
  classes |= {a['category_id'] for a in val['annotations']}

  id_map = {}

  for tfds_id, lvis_id in enumerate(
      sorted([merge_map.get(x, x) for x in classes])
  ):
    id_map[lvis_id] = tfds_id

  return id_map


def _preprocess_annotations(
    annotations_file: str, id_map: Dict[int, int]
) -> NestedDict:
  """Preprocesses the data to group together some category labels."""
  with epath.Path(annotations_file).open('r') as f:
    annotations = json.load(f)
  for ann in annotations['annotations'] + annotations['tracks']:
    ann['category_id'] = id_map[ann['category_id']]

  for ann in annotations['videos']:
    filtered_neg_ids = []
    for neg_cat_id in ann['neg_category_ids']:
      if neg_cat_id in id_map:
        filtered_neg_ids.append(id_map[neg_cat_id])
    ann['neg_category_ids'] = filtered_neg_ids
    filtered_not_exhaustive_ids = []

    for not_exhaustive_id in ann['not_exhaustive_category_ids']:
      if not_exhaustive_id in id_map:
        filtered_not_exhaustive_ids.append(id_map[not_exhaustive_id])
    ann['not_exhaustive_category_ids'] = filtered_not_exhaustive_ids
  # The frame index in the image starts at an arbitrary number. Make it
  # zero indexed.
  vid_to_frame_start = collections.defaultdict(lambda: np.inf)
  for image in annotations['images']:
    vid = image['video_id']
    vid_to_frame_start[vid] = min(vid_to_frame_start[vid], image['frame_index'])
  for image in annotations['images']:
    image['frame_index'] -= vid_to_frame_start[image['video_id']]
  return annotations


def _create_per_track_annotation(
    track,
    track_to_anns: NestedDict,
    anns_to_image: NestedDict,
    height: int,
    width: int,
) -> NestedDict:
  """Prepares annotation for a single track within a video."""
  per_track_anno = {}
  per_track_anno['bboxes'] = []
  per_track_anno['frames'] = []
  track_annos = track_to_anns[track['id']]
  assert len(np.unique([a['category_id'] for a in track_annos])) == 1
  assert len(np.unique([a['scale_category'] for a in track_annos])) == 1
  assert len(np.unique([a['iscrowd'] for a in track_annos])) == 1
  per_track_anno['category'] = track['category_id']
  per_track_anno['is_crowd'] = track_annos[0]['iscrowd']
  per_track_anno['scale_category'] = track_annos[0]['scale_category']
  per_track_anno['track_id'] = track['id']
  for ann in track_to_anns[track['id']]:
    # NOTE: Some bbox annotations extend off the boundary of the image.
    # Below we clip them to lie within the image boundaries.
    ymin = max(0.0, ann['bbox'][1] / height)
    ymax = min(1.0, (ann['bbox'][1] + ann['bbox'][3]) / height)
    xmin = max(0.0, ann['bbox'][0] / width)
    xmax = min(1.0, (ann['bbox'][0] + ann['bbox'][2]) / width)
    per_track_anno['bboxes'].append(
        tfds.features.BBox(ymin=ymin, ymax=ymax, xmin=xmin, xmax=xmax)
    )
    per_track_anno['frames'].append(anns_to_image[ann['id']]['frame_index'])

  # Frame indices should be sorted.
  assert all(
      per_track_anno['frames'][i - 1] <= per_track_anno['frames'][i]
      for i in range(1, len(per_track_anno['frames']))
  )
  return per_track_anno


class TaoConfig(tfds.core.BuilderConfig):
  """Configuration for Tao video dataset."""

  def __init__(
      self,
      *,
      height: Optional[int] = None,
      width: Optional[int] = None,
      **kwargs,
  ):
    """The parameters specifying how the dataset will be processed.

    This allows the option to preprocess the images to a smaller fixed
    resolution. If height and width are left as None, the native size
    will be used.

    Args:
      height: optional height to resize all images to
      width: optional width to resize all images to
      **kwargs: Passed on to the constructor of `BuilderConfig`.
    """
    super(TaoConfig, self).__init__(**kwargs)
    if height is not None or width is not None:
      if height is None or width is None:
        raise ValueError('Provide either both height and width or None.')
    self.height = height
    self.width = width


class Tao(tfds.core.BeamBasedBuilder):
  """DatasetBuilder for tao dataset."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Some TAO files (HVACS and AVA videos) must be manually downloaded because
  a login to MOT is required. Please download and those data following
  the instructions at https://motchallenge.net/tao_download.php

  Download this data and move the resulting .zip files to
  ~/tensorflow_datasets/downloads/manual/

  If the data requiring manual download is not present, it will be skipped over
  and only the data not requiring manual download will be used.
  """
  BUILDER_CONFIGS = [
      TaoConfig(
          name='480_640',
          description='All images are bilinearly resized to 480 X 640',
          height=480,
          width=640,
      ),
      TaoConfig(
          name='full_resolution',
          description='The full resolution version of the dataset.',
          height=None,
          width=None,
      ),
  ]
  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    names_file = tfds.core.tfds_path('video/tao/labels.txt')
    video_shape = (
        None,
        self.builder_config.height,
        self.builder_config.width,
        3,
    )
    all_features = {
        'video': tfds.features.Video(video_shape),  # pytype: disable=wrong-arg-types  # gen-stub-imports
        'metadata': {
            'height': np.int32,
            'width': np.int32,
            'num_frames': np.int32,
            'video_name': np.str_,
            'neg_category_ids': tfds.features.Tensor(
                shape=(None,), dtype=np.int32
            ),
            'not_exhaustive_category_ids': tfds.features.Tensor(
                shape=(None,), dtype=np.int32
            ),
            'dataset': np.str_,
        },
        'tracks': tfds.features.Sequence({
            'bboxes': tfds.features.Sequence(tfds.features.BBoxFeature()),
            'category': tfds.features.ClassLabel(names_file=names_file),
            'is_crowd': np.bool_,
            'track_id': np.int32,
            'scale_category': np.str_,
            # Labels do not occur for all frames. This indicates the
            # indices of the frames that have labels.
            'frames': tfds.features.Sequence(np.int32),
        }),
    }

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(all_features),
        supervised_keys=None,
        homepage='https://taodataset.org/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    data = dl_manager.download_and_extract({
        'train': _VIDEO_URL + '1-TAO_TRAIN.zip',
        'val': _VIDEO_URL + '2-TAO_VAL.zip',
        'annotations': _ANNOTATIONS_URL,
    })

    manual_train, manual_val = _maybe_prepare_manual_data(dl_manager)
    id_map = _get_category_id_map(data['annotations'] / 'annotations-1.2')

    return {
        tfds.Split.TRAIN: self._generate_examples(
            data_path=data['train'],
            manual_path=manual_train,
            annotations_path=data['annotations']
            / 'annotations-1.2'
            / 'train.json',
            id_map=id_map,
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            data_path=data['val'],
            manual_path=manual_val,
            annotations_path=data['annotations']
            / 'annotations-1.2'
            / 'validation.json',
            id_map=id_map,
        ),
    }

  def _maybe_resize_video(self, frames_list):
    """Resizes the video depending on the build_config."""
    if self.builder_config.height is None:
      return frames_list  # Don't waste compute loading and resizing.
    resized_images = []
    cv2 = tfds.core.lazy_imports.cv2
    for frame in frames_list:
      with epath.Path(frame).open('rb') as f:
        image = tfds.core.lazy_imports.PIL_Image.open(f).convert('RGB')
        image = np.asarray(image)
      image = cv2.resize(
          image, (self.builder_config.width, self.builder_config.height)
      )
      resized_images.append(image)
    return resized_images

  def _create_metadata(
      self, video_ann: NestedDict, num_frames: int
  ) -> NestedDict:
    """Creates the metadata object for each video data example."""
    metadata = {}
    metadata['num_frames'] = num_frames
    metadata['height'] = self.builder_config.height or video_ann['height']
    metadata['width'] = self.builder_config.width or video_ann['width']
    metadata['neg_category_ids'] = video_ann['neg_category_ids']
    metadata['not_exhaustive_category_ids'] = video_ann[
        'not_exhaustive_category_ids'
    ]
    metadata['dataset'] = video_ann['metadata']['dataset']
    metadata['video_name'] = video_ann['name']
    return metadata

  def _generate_examples(
      self, data_path, manual_path, annotations_path, id_map
  ):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam
    annotations = _preprocess_annotations(annotations_path, id_map)  # pytype: disable=wrong-arg-types  # always-use-return-annotations
    outs = _build_annotations_index(annotations)
    vids, ann_to_images, track_to_anns, vid_to_tracks = outs

    def _process_example(video_id_and_path):
      """Generate a data example for a single video."""
      # The video_id_and_path are passed together because beam expects
      # a function of a single param (see usage below).
      video_id, path = video_id_and_path
      video_ann = vids[video_id]
      data_example = {}
      frames = self._maybe_resize_video(list(path.iterdir()))
      data_example['video'] = frames
      data_example['metadata'] = self._create_metadata(video_ann, len(frames))
      data_example['tracks'] = []

      for track in vid_to_tracks[video_id]:
        data_example['tracks'].append(
            _create_per_track_annotation(
                track,
                track_to_anns,
                ann_to_images,
                video_ann['height'],
                video_ann['width'],
            )
        )
      return video_ann['name'], data_example

    filtered_ids = []
    for video_id, video_ann in list(vids.items()):
      # These must be manually downloaded.
      is_manual = (
          'HACS' in video_ann['metadata']['dataset']
          or 'AVA' in video_ann['metadata']['dataset']
      )
      if is_manual and manual_path is None:
        continue
      path = (
          (manual_path if is_manual else data_path)
          / 'frames'
          / vids[video_id]['name']
      )
      filtered_ids.append((video_id, path))
    return beam.Create(filtered_ids) | beam.Map(_process_example)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""https://www.kaggle.com/c/diabetic-retinopathy-detection/data."""

import csv
import io
import os

from absl import logging
from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@ONLINE {kaggle-diabetic-retinopathy,
    author = "Kaggle and EyePacs",
    title  = "Kaggle Diabetic Retinopathy Detection",
    month  = "jul",
    year   = "2015",
    url    = "https://www.kaggle.com/c/diabetic-retinopathy-detection/data"
}
"""
_URL_TEST_LABELS = "https://storage.googleapis.com/kaggle-forum-message-attachments/90528/2877/retinopathy_solution.csv"
_BTGRAHAM_DESCRIPTION_PATTERN = (
    "Images have been preprocessed as the winner of the Kaggle competition did "
    "in 2015: first they are resized so that the radius of an eyeball is "
    "{} pixels, then they are cropped to 90% of the radius, and finally they "
    "are encoded with 72 JPEG quality."
)


class DiabeticRetinopathyDetectionConfig(tfds.core.BuilderConfig):
  """BuilderConfig for DiabeticRetinopathyDetection."""

  def __init__(self, target_pixels=None, **kwargs):
    """BuilderConfig for DiabeticRetinopathyDetection.

    Args:
      target_pixels: If given, rescale the images so that the total number of
        pixels is roughly this value.
      **kwargs: keyword arguments forward to super.
    """
    super(DiabeticRetinopathyDetectionConfig, self).__init__(
        version=tfds.core.Version("3.0.0"),
        release_notes={
            "3.0.0": "New split API (https://tensorflow.org/datasets/splits)",
        },
        **kwargs,
    )
    self._target_pixels = target_pixels

  @property
  def target_pixels(self):
    return self._target_pixels


class DiabeticRetinopathyDetection(tfds.core.GeneratorBasedBuilder):
  """Diabetic retinopathy detection."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  You have to download this dataset from Kaggle.
  https://www.kaggle.com/c/diabetic-retinopathy-detection/data
  After downloading, unpack the test.zip file into test/ directory in manual_dir
  and sample.zip to sample/. Also unpack the sampleSubmissions.csv and
  trainLabels.csv.
  """

  BUILDER_CONFIGS = [
      DiabeticRetinopathyDetectionConfig(
          name="original",
          description="Images at their original resolution and quality.",
      ),
      DiabeticRetinopathyDetectionConfig(
          name="1M",
          description="Images have roughly 1,000,000 pixels, at 72 quality.",
          target_pixels=1000000,
      ),
      DiabeticRetinopathyDetectionConfig(
          name="250K",
          description="Images have roughly 250,000 pixels, at 72 quality.",
          target_pixels=250000,
      ),
      DiabeticRetinopathyDetectionConfig(
          name="btgraham-300",
          description=_BTGRAHAM_DESCRIPTION_PATTERN.format(300),
          target_pixels=300,
      ),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=(
            "A large set of high-resolution retina images taken under "
            "a variety of imaging conditions."
        ),
        features=tfds.features.FeaturesDict({
            "name": tfds.features.Text(),  # patient ID + eye. eg: "4_left".
            "image": tfds.features.Image(),
            # From 0 (no DR - saine) to 4 (Proliferative DR). -1 means no label.
            "label": tfds.features.ClassLabel(num_classes=5),
        }),
        homepage="https://www.kaggle.com/c/diabetic-retinopathy-detection/data",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    # TODO(pierrot): implement download using kaggle API.
    # TODO(pierrot): implement extraction of multiple files archives.
    path = dl_manager.manual_dir
    test_labels_path = dl_manager.download(_URL_TEST_LABELS)
    if tf.io.gfile.isdir(test_labels_path):
      # While testing: download() returns the dir containing the tests files.
      test_labels_path = os.path.join(
          test_labels_path, "retinopathy_solution.csv"
      )
    return [
        tfds.core.SplitGenerator(
            name="sample",  # 10 images, to do quicktests using dataset.
            gen_kwargs={
                "images_dir_path": os.path.join(path, "sample"),
            },
        ),
        tfds.core.SplitGenerator(
            name="train",
            gen_kwargs={
                "images_dir_path": os.path.join(path, "train"),
                "csv_path": os.path.join(path, "trainLabels.csv"),
                # CSV of the train split does not have the "Usage" column.
                # 35,126 examples.
                "csv_usage": None,
            },
        ),
        tfds.core.SplitGenerator(
            name="validation",
            gen_kwargs={
                "images_dir_path": os.path.join(path, "test"),
                "csv_path": test_labels_path,
                # Validation split corresponds to the public leaderboard data.
                # 10,906 examples.
                "csv_usage": "Public",
            },
        ),
        tfds.core.SplitGenerator(
            name="test",
            gen_kwargs={
                "images_dir_path": os.path.join(path, "test"),
                "csv_path": test_labels_path,
                # Test split corresponds to the public leaderboard data.
                # 42,670 examples.
                "csv_usage": "Private",
            },
        ),
    ]

  def _generate_examples(self, images_dir_path, csv_path=None, csv_usage=None):
    """Yields Example instances from given CSV.

    Args:
      images_dir_path: path to dir in which images are stored.
      csv_path: optional, path to csv file with two columns: name of image and
        label. If not provided, just scan image directory, don't set labels.
      csv_usage: optional, subset of examples from the csv file to use based on
        the "Usage" column from the csv.
    """
    if csv_path:
      with epath.Path(csv_path).open() as csv_f:
        reader = csv.DictReader(csv_f)
        data = [
            (row["image"], int(row["level"]))
            for row in reader
            if csv_usage is None or row["Usage"] == csv_usage
        ]
    else:
      data = [
          (fname[:-5], -1)
          for fname in tf.io.gfile.listdir(images_dir_path)
          if fname.endswith(".jpeg")
      ]
    for name, label in data:
      image_filepath = "%s/%s.jpeg" % (images_dir_path, name)
      record = {
          "name": name,
          "image": self._process_image(image_filepath),
          "label": label,
      }
      yield name, record

  def _process_image(self, filepath):
    with tf.io.gfile.GFile(filepath, mode="rb") as image_fobj:
      if self.builder_config.name.startswith("btgraham"):
        return _btgraham_processing(
            image_fobj=image_fobj,
            filepath=filepath,
            target_pixels=self.builder_config.target_pixels,
            crop_to_radius=True,
        )
      else:
        return _resize_image_if_necessary(
            image_fobj=image_fobj,
            target_pixels=self.builder_config.target_pixels,
        )


def _resize_image_if_necessary(image_fobj, target_pixels=None):
  """Resize an image to have (roughly) the given number of target pixels.

  Args:
    image_fobj: File object containing the original image.
    target_pixels: If given, number of pixels that the image must have.

  Returns:
    A file object.
  """
  if target_pixels is None:
    return image_fobj

  cv2 = tfds.core.lazy_imports.cv2
  # Decode image using OpenCV2.
  image = cv2.imdecode(
      np.frombuffer(image_fobj.read(), dtype=np.uint8), flags=3
  )
  # Get image height and width.
  height, width, _ = image.shape
  actual_pixels = height * width
  if actual_pixels > target_pixels:
    factor = np.sqrt(target_pixels / actual_pixels)
    image = cv2.resize(image, dsize=None, fx=factor, fy=factor)
  # Encode the image with quality=72 and store it in a BytesIO object.
  _, buff = cv2.imencode(".jpg", image, [int(cv2.IMWRITE_JPEG_QUALITY), 72])
  return io.BytesIO(buff.tobytes())


def _btgraham_processing(
    image_fobj, filepath, target_pixels, crop_to_radius=False
):
  """Process an image as the winner of the 2015 Kaggle competition.

  Args:
    image_fobj: File object containing the original image.
    filepath: Filepath of the image, for logging purposes only.
    target_pixels: The number of target pixels for the radius of the image.
    crop_to_radius: If True, crop the borders of the image to remove gray areas.

  Returns:
    A file object.
  """
  cv2 = tfds.core.lazy_imports.cv2
  # Decode image using OpenCV2.
  image = cv2.imdecode(
      np.frombuffer(image_fobj.read(), dtype=np.uint8), flags=3
  )
  # Process the image.
  image = _scale_radius_size(image, filepath, target_radius_size=target_pixels)
  image = _subtract_local_average(image, target_radius_size=target_pixels)
  image = _mask_and_crop_to_radius(
      image,
      target_radius_size=target_pixels,
      radius_mask_ratio=0.9,
      crop_to_radius=crop_to_radius,
  )
  # Encode the image with quality=72 and store it in a BytesIO object.
  _, buff = cv2.imencode(".jpg", image, [int(cv2.IMWRITE_JPEG_QUALITY), 72])
  return io.BytesIO(buff.tobytes())


def _scale_radius_size(image, filepath, target_radius_size):
  """Scale the input image so that the radius of the eyeball is the given."""
  cv2 = tfds.core.lazy_imports.cv2
  x = image[image.shape[0] // 2, :, :].sum(axis=1)
  r = (x > x.mean() / 10.0).sum() / 2.0
  if r < 1.0:
    # Some images in the dataset are corrupted, causing the radius heuristic to
    # fail. In these cases, just assume that the radius is the height of the
    # original image.
    logging.info('Radius of image "%s" could not be determined.', filepath)
    r = image.shape[0] / 2.0
  s = target_radius_size / r
  return cv2.resize(image, dsize=None, fx=s, fy=s)


def _subtract_local_average(image, target_radius_size):
  cv2 = tfds.core.lazy_imports.cv2
  image_blurred = cv2.GaussianBlur(image, (0, 0), target_radius_size / 30)
  image = cv2.addWeighted(image, 4, image_blurred, -4, 128)
  return image


def _mask_and_crop_to_radius(
    image, target_radius_size, radius_mask_ratio=0.9, crop_to_radius=False
):
  """Mask and crop image to the given radius ratio."""
  cv2 = tfds.core.lazy_imports.cv2
  mask = np.zeros(image.shape)
  center = (image.shape[1] // 2, image.shape[0] // 2)
  radius = int(target_radius_size * radius_mask_ratio)
  cv2.circle(mask, center=center, radius=radius, color=(1, 1, 1), thickness=-1)
  image = image * mask + (1 - mask) * 128
  if crop_to_radius:
    x_max = min(image.shape[1] // 2 + radius, image.shape[1])
    x_min = max(image.shape[1] // 2 - radius, 0)
    y_max = min(image.shape[0] // 2 + radius, image.shape[0])
    y_min = max(image.shape[0] // 2 - radius, 0)
    image = image[y_min:y_max, x_min:x_max, :]
  return image



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.image_feature."""

import contextlib
import enum
import os
import pathlib
from typing import Optional, Type

from absl.testing import parameterized
import numpy as np
import pytest
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib

randint = np.random.randint


class LibWithImportError(enum.Enum):
  CV2 = 'cv2'
  PIL = 'pil'


@contextlib.contextmanager
def make_cv2_fail():
  """Recreates OpenCV not being installed during the test."""
  real_cv2 = tfds.features.image_feature.cv2
  tfds.features.image_feature.cv2 = None
  try:
    yield LibWithImportError.CV2
  finally:
    tfds.features.image_feature.cv2 = real_cv2


@contextlib.contextmanager
def make_pil_fail():
  """Recreates Pillow not being installed during the test."""
  real_pil = tfds.features.image_feature.PIL_Image
  tfds.features.image_feature.PIL_Image = None
  try:
    yield LibWithImportError.PIL
  finally:
    tfds.features.image_feature.PIL_Image = real_pil


@contextlib.contextmanager
def make_none_fail():
  """Used for a regular scenario, where both OpenCV and PIL are installed."""
  yield


def _unsupported_images_for_pil(
    np_dtype: Type[np.generic], failing_lib: Optional[LibWithImportError]
) -> bool:
  # PIL is used when OpenCV is not installed, but PIL doesn't support 16-bit
  # images.
  return np_dtype == np.uint16 and failing_lib == LibWithImportError.CV2


class ImageFeatureTest(
    testing.FeatureExpectationsTestCase, parameterized.TestCase
):

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtypes=[
          (np.uint8, np.uint8),
          (np.uint16, np.uint16),
          (tf.uint8, np.uint8),
          (tf.uint16, np.uint16),
      ],
      channels=[1, 3, 4],
  )
  def test_images(self, make_lib_fail, dtypes, channels):
    dtype, np_dtype = dtypes
    with make_lib_fail() as failing_lib:
      if _unsupported_images_for_pil(np_dtype, failing_lib):
        return
      img = randint(256, size=(128, 100, channels), dtype=np_dtype)
      img_other_shape = randint(256, size=(64, 200, channels), dtype=np_dtype)

      filename = {
          np.uint8: {
              1: '6pixels_grayscale.png',
              3: '6pixels.png',
              4: '6pixels_4chan.png',
          },
          np.uint16: {
              1: '6pixels_grayscale_16bit.png',
              3: '6pixels_16bit.png',
              4: '6pixels_16bit_4chan.png',
          },
      }[np_dtype][channels]

      img_file_path = os.path.join(
          os.path.dirname(__file__), '../../testing/test_data', filename
      )
      with tf.io.gfile.GFile(img_file_path, 'rb') as f:
        img_byte_content = f.read()
      img_file_expected_content = np.array(
          [  # see tensorflow_datasets/testing/test_data/README.md
              [[0, 255, 0, 255], [255, 0, 0, 255], [255, 0, 255, 255]],
              [[0, 0, 255, 255], [255, 255, 0, 255], [126, 127, 128, 255]],
          ],
          dtype=np_dtype,
      )[
          :, :, :channels
      ]  # Truncate (h, w, 4) -> (h, w, c)
      if dtype == np.uint16 or dtype == tf.uint16:
        img_file_expected_content *= 257  # Scale int16 images

      numpy_array = testing.FeatureExpectationItem(
          value=img,
          expected=img,
          expected_np=img,
      )
      file_path_as_string = testing.FeatureExpectationItem(
          value=img_file_path,
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      file_path_as_path = testing.FeatureExpectationItem(
          value=pathlib.Path(img_file_path),
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      images_bytes = testing.FeatureExpectationItem(
          value=img_byte_content,
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      img_shape_can_be_dynamic = testing.FeatureExpectationItem(
          value=img_other_shape,
          expected=img_other_shape,
          expected_np=img_other_shape,
      )
      invalid_type = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128, channels), dtype=np.uint32),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='dtype should be',
      )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, channels), dtype=dtype),
          shape=(None, None, channels),
          dtype=dtype,
          tests=[
              numpy_array,
              file_path_as_string,
              file_path_as_path,
              images_bytes,
              img_shape_can_be_dynamic,
              invalid_type,
          ],
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtypes=[
          (np.uint8, np.uint8),
          (tf.uint8, np.uint8),
          (np.uint16, np.uint16),
          (tf.uint16, np.uint16),
      ],
      channels=[3, 4],
  )
  def test_images_with_invalid_shape(self, make_lib_fail, dtypes, channels):
    dtype, np_dtype = dtypes
    with make_lib_fail() as failing_lib:
      if _unsupported_images_for_pil(np_dtype, failing_lib):
        return
      invalid_number_of_dimensions = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128), dtype=np_dtype),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='must have the same rank',
      )
      invalid_number_of_channels = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128, 1), dtype=np_dtype),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='are incompatible',
      )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, channels), dtype=dtype),
          shape=(None, None, channels),
          dtype=dtype,
          tests=[invalid_number_of_dimensions, invalid_number_of_channels],
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail]
  )
  def test_image_with_statically_defined_shape(self, make_lib_fail):
    with make_lib_fail():
      img_shaped = randint(256, size=(32, 64, 1), dtype=np.uint8)

      img_shape_should_be_conserved = testing.FeatureExpectationItem(
          value=img_shaped,
          expected=img_shaped,
          expected_np=img_shaped,
      )
      img_shape_should_be_static = testing.FeatureExpectationItem(
          value=randint(256, size=(31, 64, 1), dtype=np.uint8),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='are incompatible',
      )
      self.assertFeature(
          feature=features_lib.Image(
              shape=(32, 64, 1),
              encoding_format='png',
              use_colormap=True,
          ),
          shape=(32, 64, 1),
          dtype=np.uint8,
          tests=[img_shape_should_be_conserved, img_shape_should_be_static],
          test_attributes=dict(
              _encoding_format='png',
              _use_colormap=True,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtype=[np.float32, tf.float32],
  )
  def test_images_float(self, make_lib_fail, dtype):
    with make_lib_fail():
      img = np.random.rand(28, 28, 1).astype(np.float32)
      img_other_shape = np.random.rand(12, 34, 1).astype(np.float32)

      numpy_array = testing.FeatureExpectationItem(
          value=img,
          expected=img,
          expected_np=img,
      )
      img_shape_can_be_dynamic = testing.FeatureExpectationItem(
          value=img_other_shape,
          expected=img_other_shape,
          expected_np=img_other_shape,
      )
      invalid_type = testing.FeatureExpectationItem(
          value=img.astype(np.float64),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='dtype should be',
      )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, 1), dtype=dtype),
          shape=(None, None, 1),
          dtype=dtype,
          tests=[
              numpy_array,
              img_shape_can_be_dynamic,
              invalid_type,
          ],
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      dtypes=[
          (np.uint16, np.uint16),
          (tf.uint16, np.uint16),
      ],
      channels=[1, 3, 4],
  )
  def test_cannot_decode_16_bit_images_with_pil(self, dtypes, channels):
    dtype, np_dtype = dtypes
    # PIL is used when OpenCV fails.
    with make_cv2_fail():
      self.assertFeature(
          feature=features_lib.Image(shape=(128, 128, channels), dtype=dtype),
          shape=(128, 128, channels),
          dtype=dtype,
          tests=[
              testing.FeatureExpectationItem(
                  value=randint(256, size=(128, 128, channels), dtype=np_dtype),
                  raise_cls_np=ValueError,
                  raise_msg='PIL does not handle',
              )
          ],
      )

  def test_fail_if_neither_opencv_nor_pil_is_installed(self):
    with make_pil_fail():
      with make_cv2_fail():
        self.assertFeature(
            feature=features_lib.Image(shape=(128, 128, 3), dtype=np.uint8),
            shape=(128, 128, 3),
            dtype=np.uint8,
            tests=[
                testing.FeatureExpectationItem(
                    value=randint(256, size=(128, 128, 3), dtype=np.uint8),
                    raise_cls_np=ImportError,
                    raise_msg='requires either OpenCV or PIL',
                )
            ],
        )


@pytest.mark.parametrize(
    'shape, dtype, encoding_format, err_msg',
    [
        (None, np.uint16, r'jpeg', 'Acceptable `dtype` for jpeg:'),
        (None, tf.uint16, r'jpeg', 'Acceptable `dtype` for jpeg:'),
        (None, np.float32, None, 'only support single-channel'),
        (None, tf.float32, None, 'only support single-channel'),
        ((None, None, 1), np.float64, None, 'Acceptable `dtype`'),
        ((None, None, 1), tf.float64, None, 'Acceptable `dtype`'),
    ],
)
def test_invalid_img(shape, dtype, encoding_format, err_msg):
  with pytest.raises(ValueError, match=err_msg):
    features_lib.Image(
        shape=shape, dtype=dtype, encoding_format=encoding_format
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gsm8k dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.gsm8k import gsm8k


class Gsm8kTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for gsm8k dataset."""

  DATASET_CLASS = gsm8k.Gsm8k
  SPLITS = {
      'train': 1,
      'test': 1,
      'train_socratic': 1,
      'test_socratic': 1,
  }
  DL_EXTRACT_RESULT = {
      k: f'{k}.jsonl'
      for k in ['train', 'test', 'train_socratic', 'test_socratic']
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_dmlab_rooms_select_nonmatching_object dataset."""
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import dmlab_dataset
from tensorflow_datasets.rl_unplugged.rlu_dmlab_rooms_select_nonmatching_object import rlu_dmlab_rooms_select_nonmatching_object


class RluDmlabRoomsSelectNonmatchingObjectTest(
    tfds.testing.DatasetBuilderTestCase
):
  """Tests for rlu_dmlab_explore_object_rewards_many dataset."""

  DATASET_CLASS = (
      rlu_dmlab_rooms_select_nonmatching_object.RluDmlabRoomsSelectNonmatchingObject
  )
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  # The different configs are only different in the name of the files.
  BUILDER_CONFIG_NAMES_TO_TEST = ['training_0']

  @classmethod
  def setUpClass(cls):
    dmlab_dataset.DMLabDatasetBuilder._INPUT_FILE_PREFIX = cls.dummy_data
    dmlab_dataset.DMLabDatasetBuilder._SHARDS = 1
    super().setUpClass()


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLI Fixtures."""

import argparse
from unittest import mock

import pytest


@pytest.fixture(scope='session', autouse=True)
def _mock_argparse_flags():
  """Mock absl argparse_flag with argparse."""
  # absl flags are globally registered during pytest tests collection.
  # This creates conflict if the same flag is defined outside the CLI (in
  # another test):
  # `flags.DEFINE_string('data_dir')` with `parser.add_argument('--data_dir')`
  # We patch argparse_flags during test, so absl flags are ignored.
  with mock.patch(
      'absl.flags.argparse_flags.ArgumentParser', argparse.ArgumentParser
  ):
    yield



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""WikipediaToxicitySubtypes from Jigsaw Toxic Comment Classification Challenge."""

import csv
import os

from etils import epath
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{10.1145/3038912.3052591,
  author = {Wulczyn, Ellery and Thain, Nithum and Dixon, Lucas},
  title = {Ex Machina: Personal Attacks Seen at Scale},
  year = {2017},
  isbn = {9781450349130},
  publisher = {International World Wide Web Conferences Steering Committee},
  address = {Republic and Canton of Geneva, CHE},
  url = {https://doi.org/10.1145/3038912.3052591},
  doi = {10.1145/3038912.3052591},
  booktitle = {Proceedings of the 26th International Conference on World Wide Web},
  pages = {1391-1399},
  numpages = {9},
  keywords = {online discussions, wikipedia, online harassment},
  location = {Perth, Australia},
  series = {WWW '17}
}
"""

_SUBTYPES_HOMEPAGE = 'https://www.kaggle.com/c/jigsaw-toxic-comment-classification-challenge/data'
_MULTILINGUAL_HOMEPAGE = 'https://www.kaggle.com/c/jigsaw-multilingual-toxic-comment-classification/data'

_COMMON_DESCRIPTION = """
The comments in this dataset come from an archive of Wikipedia talk page
comments. These have been annotated by Jigsaw for toxicity, as well as (for the
main config) a variety of toxicity subtypes, including severe toxicity,
obscenity, threatening language, insulting language, and identity attacks. This
dataset is a replica of the data released for the Jigsaw Toxic Comment
Classification Challenge and Jigsaw Multilingual Toxic Comment Classification
competition on Kaggle, with the test dataset merged with the test_labels
released after the end of the competitions. Test data not used for scoring has
been dropped. This dataset is released under CC0, as is the underlying comment
text.
"""

_SUBTYPES_DESCRIPTION = """
The comments in the WikipediaToxicitySubtypes config are from an archive of
English Wikipedia talk page comments which have been annotated by Jigsaw for
toxicity, as well as five toxicity subtype labels (severe toxicity, obscene,
threat, insult, identity_attack). The toxicity and toxicity subtype labels are
binary values (0 or 1) indicating whether the majority of annotators assigned
that attribute to the comment text. This config is a replica of the data
released for the Jigsaw Toxic Comment Classification Challenge on Kaggle, with
the test dataset joined with the test_labels released after the competition, and
test data not used for scoring dropped.

See the Kaggle documentation
https://www.kaggle.com/c/jigsaw-toxic-comment-classification-challenge/data
or https://figshare.com/articles/Wikipedia_Talk_Labels_Toxicity/4563973 for more
details.
"""

_MULTILINGUAL_DESCRIPTION = """
The comments in the WikipediaToxicityMultilingual config here are from an
archive of non-English Wikipedia talk page comments annotated by Jigsaw for
toxicity, with a binary value (0 or 1) indicating whether the majority of
annotators rated the comment text as toxic. The comments in this config are in
multiple different languages (Turkish, Italian, Spanish, Portuguese, Russian,
and French). This config is a replica of the data released for the Jigsaw
Multilingual Toxic Comment Classification on Kaggle, with the test dataset
joined with the test_labels released after the competition.

See the Kaggle documentation
https://www.kaggle.com/c/jigsaw-multilingual-toxic-comment-classification/data
for more details.
"""

_DOWNLOAD_URL = 'https://storage.googleapis.com/jigsaw-unintended-bias-in-toxicity-classification/wikipedia_toxicity_subtypes_v0.3.zip'

TOXICITY_SUBTYPES = [
    'severe_toxicity',
    'obscene',
    'threat',
    'insult',
    'identity_attack',
]


class WikipediaToxicityConfig(tfds.core.BuilderConfig):
  """Configuration for `WikipediaToxicitySubtypes`."""

  def __init__(self, name: str, description: str, multilingual: bool):
    super(WikipediaToxicityConfig, self).__init__(
        name=name, description=description
    )
    self.multilingual = multilingual


class WikipediaToxicitySubtypes(tfds.core.GeneratorBasedBuilder):
  """Classification of 295K Wikipedia talk page comments for types of toxicity.

  This version of the Wikipedia Toxicity Subtypes dataset provides access to the
  primary toxicity label annotated by crowd workers, with five additional
  toxicity subtype labels in the main config. The toxicity and toxicity subtype
  labels are binary values (0 or 1) indicating whether the majority of
  annotators assigned that attributes to the comment text.

  While the main config is entirely in English, the
  WikipediaToxicityMultilingual config provides validation and test splits with
  data in several non-English languages (tr, it, es, pt, ru, fr). This config
  provides access only to the primary toxicity label for each comment.
  """

  BUILDER_CONFIGS = [
      WikipediaToxicityConfig(
          name='EnglishSubtypes',
          description=_SUBTYPES_DESCRIPTION,
          multilingual=False,
      ),
      WikipediaToxicityConfig(
          name='Multilingual',
          description=_MULTILINGUAL_DESCRIPTION,
          multilingual=True,
      ),
  ]

  VERSION = tfds.core.Version('0.3.1')
  RELEASE_NOTES = {
      '0.3.1': (
          'Added a unique id for each comment. (For the Multilingual '
          'config, these are only unique within each split.)'
      ),
      '0.3.0': 'Added WikipediaToxicityMultilingual config.',
      '0.2.0': 'Updated features for consistency with CivilComments dataset.',
  }

  def _info(self):
    description = _COMMON_DESCRIPTION
    homepage = (
        _MULTILINGUAL_HOMEPAGE
        if self.builder_config.multilingual
        else _SUBTYPES_HOMEPAGE
    )

    features = {
        'text': tfds.features.Text(),
        'id': tfds.features.Text(),
        'language': tfds.features.Text(),
    }
    labels = ['toxicity']
    if not self.builder_config.multilingual:
      labels += TOXICITY_SUBTYPES

    for label in labels:
      features[label] = tf.float32

    return tfds.core.DatasetInfo(
        builder=self,
        description=description,
        features=tfds.features.FeaturesDict(features),
        supervised_keys=('text', 'toxicity'),
        homepage=homepage,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_path = dl_manager.download_and_extract(_DOWNLOAD_URL)
    file_path_prefix = os.path.join(
        dl_path,
        'wikidata_' + 'multilingual_' * self.builder_config.multilingual,
    )

    split_generators = [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={'filename': file_path_prefix + 'train.csv'},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={'filename': file_path_prefix + 'test.csv'},
        ),
    ]

    if self.builder_config.multilingual:
      # Validation split instead of train for WikipediaToxicityMultilingual.
      split_generators[0] = tfds.core.SplitGenerator(
          name=tfds.Split.VALIDATION,
          gen_kwargs={'filename': file_path_prefix + 'validation.csv'},
      )

    return split_generators

  def _generate_examples(self, filename):
    """Yields examples.

    Each example contains a text input and then six annotation labels.

    Args:
      filename: the path of the file to be read for this split.

    Yields:
      A dictionary of features, all floating point except the input text.
    """
    with epath.Path(filename).open() as f:
      reader = csv.DictReader(f)
      for row in reader:
        example = {}
        example['text'] = row['comment_text']
        example['id'] = row['id']
        example['language'] = row['lang']
        example['toxicity'] = float(row['toxic'])

        if not self.builder_config.multilingual:
          example['severe_toxicity'] = float(row['severe_toxic'])
          example['identity_attack'] = float(row['identity_hate'])
          for label in ['obscene', 'threat', 'insult']:
            example[label] = float(row[label])

        yield row['id'], example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NumPy utils."""

from typing import Tuple

from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils


@py_utils.memoize()
def to_np_shape(shape: type_utils.Shape, replace_int=-1) -> Tuple[int]:
  return tuple(dim if dim is not None else replace_int for dim in shape)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to use to extract archives. No business logic."""

import bz2
import concurrent.futures
import contextlib
import gzip
import multiprocessing
import os
import tarfile
import typing
from typing import Iterator, Tuple, Union
import uuid
import zipfile

from absl import logging
from etils import epath
import promise
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import resource as resource_lib
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


@utils.memoize()
def get_extractor(*args, **kwargs):
  return _Extractor(*args, **kwargs)


class ExtractError(Exception):
  """There was an error while extracting the archive."""


class UnsafeArchiveError(Exception):
  """The archive is unsafe to unpack, e.g. absolute path."""


class _Extractor(object):
  """Singleton (use `get_extractor()` module fct) to extract archives."""

  def __init__(self, max_workers=None):
    max_workers = max_workers or multiprocessing.cpu_count()
    self._executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=max_workers * 4
    )
    self._pbar_path = None

  @contextlib.contextmanager
  def tqdm(self):
    """Add a progression bar for the current extraction."""
    with utils.async_tqdm(
        total=0, desc='Extraction completed...', unit=' file', mininterval=1.0
    ) as pbar_path:
      self._pbar_path = pbar_path
      yield

  def extract(self, path, extract_method, to_path):
    """Returns `promise.Promise` => to_path."""
    path = os.fspath(path)
    to_path = os.fspath(to_path)
    if extract_method not in _EXTRACT_METHODS:
      raise ValueError('Unknown extraction method "%s".' % extract_method)
    future = self._executor.submit(
        self._extract, from_path=path, method=extract_method, to_path=to_path
    )
    return promise.Promise.resolve(future)

  def _extract(
      self,
      from_path: epath.PathLike,
      method,
      to_path: epath.PathLike,
  ) -> epath.Path:
    """Returns `to_path` once resource has been extracted there."""
    to_path_tmp = '%s%s_%s' % (
        to_path,
        constants.INCOMPLETE_SUFFIX,
        uuid.uuid4().hex,
    )
    max_length_dst_path = 0
    futures = []
    if self._pbar_path is None:
      raise ValueError("'_pbar_path' is not initialized")
    try:
      for path, handle in iter_archive(from_path, method):
        path = tf.compat.as_text(path)
        dst_path = path and os.path.join(to_path_tmp, path) or to_path_tmp
        max_length_dst_path = max(max_length_dst_path, len(dst_path))
        file_utils.makedirs_cached(os.path.dirname(dst_path))
        future = self._executor.submit(_copy, handle.read(), dst_path)
        futures.append(future)
        self._pbar_path.update_total(1)

      # Wait until all copies have completed.
      for future in concurrent.futures.as_completed(futures):
        future.result()
        self._pbar_path.update(1)
    except BaseException as err:
      msg = f'Error while extracting {from_path} to {to_path}: {err}'
      # Check if running on windows
      if os.name == 'nt' and max_length_dst_path > 250:
        msg += (
            '\nOn windows, path lengths greater than 260 characters may result'
            ' in an error. See the doc to remove the limitation: '
            'https://docs.python.org/3/using/windows.html#removing-the-max-path-limitation'
        )
      raise ExtractError(msg)

    # `tf.io.gfile.Rename(overwrite=True)` doesn't work for non empty
    # directories, so delete destination first, if it already exists.
    to_path = epath.Path(to_path)
    if to_path.exists():
      # `rename` is atomic, but not `rmtree`.
      # When 2 builder scripts (for each config) extract the same file, one can
      # `rename` while the other is still running `rmtree`, leading to corrupted
      # archive dir.
      path_to_delete = epath.Path(to_path_tmp + '.todelete')
      to_path.rename(path_to_delete)
      path_to_delete.rmtree()
    epath.Path(to_path_tmp).rename(to_path)
    return to_path


def _copy(src_data, dest_path):
  """Copy data read from src data to new file in dest_path."""
  with tf.io.gfile.GFile(dest_path, 'wb') as dest_file:
    dest_file.write(src_data)


def _normpath(path):
  path = os.path.normpath(path)
  if (
      path.startswith('.')
      or os.path.isabs(path)
      or path.endswith('~')
      or os.path.basename(path).startswith('.')
  ):
    return None
  return path


@contextlib.contextmanager
def _open_or_pass(path_or_fobj):
  if isinstance(path_or_fobj, epath.PathLikeCls):
    with tf.io.gfile.GFile(path_or_fobj, 'rb') as f_obj:
      yield f_obj
  else:
    yield path_or_fobj


def iter_tar(arch_f, stream=False):
  """Iter over tar archive, yielding (path, object-like) tuples.

  Args:
    arch_f: File object of the archive to iterate.
    stream: If True, open the archive in stream mode which allows for faster
      processing and less temporary disk consumption, but random access to the
      file is not allowed.

  Yields:
    (filepath, extracted_fobj) for each file in the archive.
  """
  read_type = 'r' + ('|' if stream else ':') + '*'

  with _open_or_pass(arch_f) as fobj:
    tar = tarfile.open(mode=read_type, fileobj=fobj)
    for member in tar:
      if stream and (member.islnk() or member.issym()):
        # Links cannot be dereferenced in stream mode.
        logging.warning('Skipping link during extraction: %s', member.name)
        continue

      try:
        extract_file = tar.extractfile(member)
      except KeyError:
        if not (member.islnk() or member.issym()):
          raise  # Forward exception non-link files which couldn't be extracted
        # The link could not be extracted, which likely indicates a corrupted
        # archive.
        logging.warning(
            'Skipping extraction of invalid link: %s -> %s',
            member.name,
            member.linkname,
        )
        continue

      if extract_file:  # File with data (not directory):
        path = _normpath(member.path)  # pytype: disable=attribute-error
        if not path:
          continue
        yield (path, extract_file)


def iter_tar_stream(arch_f):
  return iter_tar(arch_f, stream=True)


def iter_gzip(arch_f):
  with _open_or_pass(arch_f) as fobj:
    gzip_ = gzip.GzipFile(fileobj=fobj)
    yield ('', gzip_)  # No inner file.


def iter_bzip2(arch_f):
  with _open_or_pass(arch_f) as fobj:
    bz2_ = bz2.BZ2File(filename=fobj)
    yield ('', bz2_)  # No inner file.


def iter_zip(arch_f):
  """Iterate over zip archive."""
  with _open_or_pass(arch_f) as fobj:
    z = zipfile.ZipFile(fobj)
    for member in z.infolist():
      if member.is_dir():  # Filter directories  # pytype: disable=attribute-error
        continue
      extract_file = z.open(member)
      path = _normpath(member.filename)
      if not path:
        continue
      yield (path, extract_file)


_EXTRACT_METHODS = {
    resource_lib.ExtractMethod.BZIP2: iter_bzip2,
    resource_lib.ExtractMethod.GZIP: iter_gzip,
    resource_lib.ExtractMethod.TAR: iter_tar,
    resource_lib.ExtractMethod.TAR_GZ: iter_tar,
    resource_lib.ExtractMethod.TAR_GZ_STREAM: iter_tar_stream,
    resource_lib.ExtractMethod.TAR_STREAM: iter_tar_stream,
    resource_lib.ExtractMethod.ZIP: iter_zip,
}


def iter_archive(
    path: Union[epath.PathLike, typing.BinaryIO],
    method: resource_lib.ExtractMethod,
) -> Iterator[Tuple[str, typing.BinaryIO]]:
  """Iterate over an archive.

  Args:
    path: `str`, archive path
    method: `tfds.download.ExtractMethod`, extraction method

  Returns:
    An iterator of `(path_in_archive, f_obj)`
  """
  if method == resource_lib.ExtractMethod.NO_EXTRACT:
    raise ValueError(
        f'Cannot `iter_archive` over {path}. Invalid or unrecognised archive.'
    )
  return _EXTRACT_METHODS[method](path)  # pytype: disable=bad-return-type



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_dmlab_explore_object_rewards_many dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import dmlab_dataset

_TASK = 'explore_object_rewards_many'
_EPISODE_LENGTH = 1801


class RluDmlabExploreObjectRewardsMany(dmlab_dataset.DMLabDatasetBuilder):
  """DatasetBuilder for rlu_dmlab_explore_object_rewards_many dataset."""

  VERSION = tfds.core.Version('1.2.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.0': 'Added is_last.',
      '1.2.0': 'BGR -> RGB fix for pixel observations.',
  }

  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [
      dmlab_dataset.BuilderConfig(
          name='training_0', task=_TASK, episode_length=_EPISODE_LENGTH
      ),
      dmlab_dataset.BuilderConfig(
          name='training_1', task=_TASK, episode_length=_EPISODE_LENGTH
      ),
      dmlab_dataset.BuilderConfig(
          name='training_2', task=_TASK, episode_length=_EPISODE_LENGTH
      ),
  ]
  # pytype: enable=wrong-keyword-args



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Eurosat remote sensing benchmarking dataset."""

import io
import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@misc{helber2017eurosat,
    title={EuroSAT: A Novel Dataset and Deep Learning Benchmark for Land Use and Land Cover Classification},
    author={Patrick Helber and Benjamin Bischke and Andreas Dengel and Damian Borth},
    year={2017},
    eprint={1709.00029},
    archivePrefix={arXiv},
    primaryClass={cs.CV}
}
"""

_DESCRIPTION = """\
EuroSAT dataset is based on Sentinel-2 satellite images covering 13 spectral
bands and consisting of 10 classes with 27000 labeled and
geo-referenced samples.

Two datasets are offered:
- rgb: Contains only the optical R, G, B frequency bands encoded as JPEG image.
- all: Contains all 13 bands in the original value range (float32).

URL: https://github.com/phelber/eurosat
"""

_LABELS = [
    'AnnualCrop',
    'Forest',
    'HerbaceousVegetation',
    'Highway',
    'Industrial',
    'Pasture',
    'PermanentCrop',
    'Residential',
    'River',
    'SeaLake',
]

_URL = 'https://github.com/phelber/eurosat'

_DATA_OPTIONS = ['rgb', 'all']


class EurosatConfig(tfds.core.BuilderConfig):
  """BuilderConfig for eurosat."""

  def __init__(self, selection=None, download_url=None, subdir=None, **kwargs):
    """Constructs a EurosatConfig.

    Args:
      selection: `str`, one of `_DATA_OPTIONS`.
      download_url: `str`, download URL to the zip file.
      subdir: `str`, subdir within the zip file.
      **kwargs: keyword arguments forwarded to super.
    """
    if selection not in _DATA_OPTIONS:
      raise ValueError('selection must be one of %s' % _DATA_OPTIONS)

    super(EurosatConfig, self).__init__(
        version=tfds.core.Version('2.0.0'), **kwargs
    )
    self.selection = selection
    self.download_url = download_url
    self.subdir = subdir


class Eurosat(tfds.core.GeneratorBasedBuilder):
  """Eurosat remote sensing benchmarking dataset."""

  BUILDER_CONFIGS = [
      EurosatConfig(
          selection='rgb',
          name='rgb',
          download_url='http://madm.dfki.de/files/sentinel/EuroSAT.zip',
          subdir='2750',
          description='Sentinel-2 RGB channels',
      ),
      EurosatConfig(
          selection='all',
          name='all',
          download_url='http://madm.dfki.de/files/sentinel/EuroSATallBands.zip',
          subdir='ds/images/remote_sensing/otherDatasets/sentinel_2/tif',
          description='13 Sentinel-2 channels',
      ),
  ]

  def _info(self):
    if self.builder_config.selection == 'rgb':
      features = tfds.features.FeaturesDict({
          'image': tfds.features.Image(shape=[64, 64, 3]),
          'label': tfds.features.ClassLabel(names=_LABELS),
          'filename': tfds.features.Text(),
      })
      supervised_keys = ('image', 'label')
    elif self.builder_config.selection == 'all':
      features = tfds.features.FeaturesDict({
          'sentinel2': tfds.features.Tensor(
              shape=[64, 64, 13], dtype=np.float32
          ),
          'label': tfds.features.ClassLabel(names=_LABELS),
          'filename': tfds.features.Text(),
      })
      supervised_keys = ('sentinel2', 'label')

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=features,
        supervised_keys=supervised_keys,
        homepage=_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(self.builder_config.download_url)
    path = os.path.join(path, self.builder_config.subdir)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'path': path,
                'selection': self.builder_config.selection,
            },
        ),
    ]

  def _generate_examples(self, path, selection):
    """Yields examples."""
    for filename in tf.io.gfile.glob(os.path.join(path, '*', '*')):
      label = os.path.split(filename)[-1].split('_')[0]
      if selection == 'rgb':
        record = {
            'image': filename,
            'label': label,
            'filename': os.path.basename(filename),
        }
      else:
        record = {
            'sentinel2': _extract_channels(filename),
            'label': label,
            'filename': os.path.basename(filename),
        }
      yield f'{label}_{os.path.basename(filename)}', record


def _extract_channels(filename):
  with tf.io.gfile.GFile(filename, 'rb') as f:
    arr = tfds.core.lazy_imports.tifffile.imread(io.BytesIO(f.read()))

  arr = arr.astype('float32')
  return arr



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for delete_old_versions.py script."""

import pathlib

from typing import List

from tensorflow_datasets.scripts.cleanup import delete_old_versions


def test_split_full_names():
  """Test _split_full_names function."""
  full_names = [
      'my_dataset/1.2.0',
      'my_dataset/1.3.0',
      'my_other_dataset/config/2.0.0',
      'my_other_dataset/config/1.3.0',
      'my_other_dataset/other_config/1.3.0',
  ]
  assert delete_old_versions._split_full_names(full_names) == {
      'my_dataset': {
          '1.2.0': {},
          '1.3.0': {},
      },
      'my_other_dataset': {
          'config': {
              '2.0.0': {},
              '1.3.0': {},
          },
          'other_config': {
              '1.3.0': {},
          },
      },
  }


def test_delete_script(tmp_path: pathlib.PurePath):
  # Create paths
  for dataset_dir in [
      'extracted',  # DownloadManager directories should be preserved.
      'dataset/config/1.0.0',
      'my_dataset/other_dir',
      'my_dataset/1.0.0/other',
      'my_dataset/1.1.0',
      'my_dataset/1.2.0/diff',
      'my_dataset/1.2.0/other',
      'my_dataset/1.3.0',
      'my_other_dataset/config/1.2.0',
      'my_other_dataset/config/1.3.0',
      'my_other_dataset/other_config/1.2.0',
      'my_other_dataset/other_config/1.3.0',
      'my_other_dataset/yet_other_config',
      'my_other_dataset/config_deprecated/1.2.0',
      'old_dataset',
  ]:
    tmp_path.joinpath(dataset_dir).mkdir(parents=True)

  dirs_to_keep, dirs_to_delete = delete_old_versions._get_extra_dirs(
      data_dir=tmp_path,
      current_full_names=[
          'non_generated_dataset0/1.0.0',
          'non_generated_dataset1/config/1.0.0',
          'dataset/config/1.0.0',
          'my_dataset/1.2.0',
          'my_dataset/1.3.0',
          'my_other_dataset/config/1.3.0',
          'my_other_dataset/other_config/1.3.0',
          'another_dataset/config/1.2.0',
          'another_dataset/other_config/1.1.0',
      ],
  )
  assert dirs_to_keep == _norm_path(
      tmp_path,
      [
          'dataset/config/1.0.0',
          'my_dataset/1.2.0',
          'my_dataset/1.3.0',
          'my_other_dataset/config/1.3.0',
          'my_other_dataset/other_config/1.3.0',
      ],
  )
  assert dirs_to_delete == _norm_path(
      tmp_path,
      [
          'my_dataset/1.0.0',
          'my_dataset/1.1.0',
          'my_dataset/other_dir',
          'my_other_dataset/config/1.2.0',
          'my_other_dataset/config_deprecated',
          'my_other_dataset/other_config/1.2.0',
          'my_other_dataset/yet_other_config',
          'old_dataset',
      ],
  )


def _norm_path(root_path: pathlib.Path, paths: List[str]) -> List[pathlib.Path]:
  """Normalize paths (for windows compatibility)."""
  return [root_path / p for p in paths]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic related to reading datasets metadata from config files."""

import dataclasses
import functools
from typing import Dict, List, Optional

from etils import epath
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import resource_utils


CITATIONS_FILENAME = "CITATIONS.bib"
DESCRIPTIONS_FILENAME = "README.md"
TAGS_FILENAME = "TAGS.txt"

_METADATA_FILES = [
    CITATIONS_FILENAME,
    DESCRIPTIONS_FILENAME,
    TAGS_FILENAME,
]


@dataclasses.dataclass(frozen=True)
class DatasetMetadata:
  """Contains Dataset metadata read from configs."""

  description: str
  citation: str
  tags: List[str]


def _get_tags(tags_txt: str) -> List[str]:
  """Returns list of tags from raw tags file content."""
  tags = []
  for line in tags_txt.split("\n"):
    tag = line.split("#", 1)[0].strip()
    if tag:
      tags.append(tag)
  return tags


def _get_valid_tags_text() -> str:
  """Returns the valid_tags.txt content."""
  path = resource_utils.tfds_path() / "core/valid_tags.txt"
  return path.read_text("utf-8")


def valid_tags() -> List[str]:
  """Returns a list of valid tags."""
  return _get_tags(_get_valid_tags_text())


def valid_tags_with_comments() -> str:
  """Returns valid tags (one per line) with comments."""
  return "\n".join(
      [
          line
          for line in _get_valid_tags_text().split("\n")
          if not line.startswith("#")
      ]
  )


@functools.lru_cache(maxsize=256)
def load(pkg_path: epath.Path) -> DatasetMetadata:
  """Returns dataset metadata loaded from files in pkg."""
  raw_metadata = _read_files(pkg_path)
  tags = _get_tags(raw_metadata.get(TAGS_FILENAME, ""))
  return DatasetMetadata(
      description=raw_metadata.get(DESCRIPTIONS_FILENAME, None),
      citation=raw_metadata.get(CITATIONS_FILENAME, None),
      tags=tags,
  )


def _read_files(path: epath.Path) -> Dict[str, str]:
  """Reads all metadata files content.

  Args:
    path: path to package where metadata files are.

  Returns:
    dict {path: content}, where path is relative to the dataset src directory.
    e.g. {'README.md': '...', 'CITATIONS.cff': '...'}
  """
  name2path = {}
  for inode in path.iterdir():
    if inode.name in _METADATA_FILES:
      name2path[inode.name] = path.joinpath(inode.name)
  return utils.tree.parallel_map(
      lambda f: f.read_text(encoding="utf-8"), name2path
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wiki_dialog dataset."""

from tensorflow_datasets import testing
from tensorflow_datasets.text.wiki_dialog import wiki_dialog


class WikiDialogTest(testing.DatasetBuilderTestCase):
  """Tests for wiki_dialog dataset."""

  DATASET_CLASS = wiki_dialog.WikiDialog
  BUILDER_CONFIG_NAMES_TO_TEST = ['OQ']
  DL_EXTRACT_RESULT = {
      'train': ['data_train.jsonl-00000-of-00001.gz'],
      'validation': ['data_validation.jsonl.gz'],
  }

  SPLITS = {
      'train': 3,  # Number of train pairs.
      'validation': 1,  # Number of validation pairs.
  }


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.text.cnn_dailymail."""

import tempfile

from tensorflow_datasets import testing
from tensorflow_datasets.summarization import cnn_dailymail

_STORY_FILE = b"""Some article.
This is some article text.

@highlight

highlight text.

@highlight

Highlight two

@highlight

highlight Three
"""


class CnnDailymailTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = cnn_dailymail.CnnDailymail
  SPLITS = {'train': 3, 'validation': 2, 'test': 2}
  DL_EXTRACT_RESULT = {
      'cnn_stories': '',
      'dm_stories': '',
      'test_urls': 'all_test.txt',
      'train_urls': 'all_train.txt',
      'val_urls': 'all_val.txt',
  }

  def test_get_art_abs(self):
    with tempfile.NamedTemporaryFile(delete=True) as f:
      f.write(_STORY_FILE)
      f.flush()
      article, abstract = cnn_dailymail._get_art_abs(f.name)
      self.assertEqual('Some article. This is some article text.', article)

      self.assertEqual(
          'highlight text.\nHighlight two.\nhighlight Three.', abstract
      )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""tao dataset."""

import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.video.tao import tao


class TaoTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for tao dataset."""

  DATASET_CLASS = tao.Tao
  SPLITS = {
      tfds.Split.TRAIN: 1,
      tfds.Split.VALIDATION: 1,
  }
  DL_EXTRACT_RESULT = {
      'train': '',
      'val': '',
      'annotations': '',
  }

  BUILDER_CONFIG_NAMES_TO_TEST = ['test_config']

  @classmethod
  def setUpClass(cls):
    tao.Tao.BUILDER_CONFIGS = [
        tao.TaoConfig(
            name='test_config',
            description='All images are bilinearly resized to 28 X 42',
            height=28,
            width=42,
        ),
    ]
    super().setUpClass()

  def _download_and_prepare_as_dataset(self, builder):
    super()._download_and_prepare_as_dataset(builder)

    if not tf.executing_eagerly():  # Only test the following in eager mode.
      return

    with self.subTest('check_annotations'):
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      for ex in [train_ex, val_ex]:
        # There should be the same number of each of these; a number
        # per group of bboxes indicating which frame they correspond to.
        self.assertEqual(
            ex['tracks']['bboxes'].shape[0], ex['tracks']['frames'].shape[0]
        )

    with self.subTest('check_video'):
      splits = builder.as_dataset()
      train_ex = list(splits[tfds.Split.TRAIN])[0]
      val_ex = list(splits[tfds.Split.VALIDATION])[0]
      # NOTE: For real images, this will be a list of potentially a thousand or
      # more frames. For testing purposes we load a single dummy 10 X 10 image.
      self.assertEqual(train_ex['video'].shape, (1, 28, 42, 3))
      self.assertEqual(val_ex['video'].shape, (1, 28, 42, 3))


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFDS DType utils to handle both NumPy and TensorFlow DTypes."""

from typing import cast, Any, List, Type, Union

from absl import logging
from etils import enp
import numpy as np
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

TfdsDType = type_utils.TfdsDType


@py_utils.memoize()
def cast_to_numpy(dtype: type_utils.TfdsDType) -> np.dtype:
  """Casts any np.dtype or tf.dtypes.DType to np.dtype.

  Args:
    dtype: input DType.

  Returns:
    Equivalent NumPy DType.
  """
  tf_already_loaded = enp.lazy.has_tf
  if tf_already_loaded and isinstance(dtype, tf.dtypes.DType):
    np_dtype = np.dtype(dtype.as_numpy_dtype)
    logging.log_first_n(
        logging.WARNING,
        (
            f'You use TensorFlow DType {dtype} in tfds.features '
            'This will soon be deprecated in favor of NumPy DTypes. '
            f'In the meantime it was converted to {np_dtype}.'
        ),
        10,
    )
    return np_dtype
  # Strings are managed as np.object_ (rather than np.str_) in order to
  # optimize for memory and for consistency with tf.string:
  # np.array(['', '1234567890'], dtype=np.str_).nbytes == 80
  # np.array(['', '1234567890'], dtype=np.object_).nbytes == 16
  if is_string(dtype):
    return np.dtype(np.object_)
  return cast(np.dtype, dtype)


def is_np_or_tf_dtype(value: Any) -> bool:
  """Returns True is the given value is a NumPy or TensorFlow dtype."""
  return enp.lazy.is_np_dtype(value) or enp.lazy.is_tf_dtype(value)


def _is_dtype(
    numpy_dtypes: List[Union[np.dtype, Type[np.generic]]],
    tf_dtype: Any,
    dtype: TfdsDType,
) -> bool:
  if enp.lazy.is_np_dtype(dtype):
    return any(
        [is_np_sub_dtype(dtype, numpy_dtype) for numpy_dtype in numpy_dtypes]
    )
  if enp.lazy.has_tf and isinstance(dtype, tf.dtypes.DType):
    if isinstance(tf_dtype, str):
      return getattr(dtype, tf_dtype)
    return dtype == tf_dtype
  raise TypeError(f'type {dtype} ({type(dtype)}) not recognized')


@py_utils.memoize()
def is_bool(dtype: TfdsDType) -> bool:
  return _is_dtype([np.bool_], 'is_bool', dtype)


@py_utils.memoize()
def is_floating(dtype: TfdsDType) -> bool:
  return _is_dtype([np.floating], 'is_floating', dtype)


@py_utils.memoize()
def is_integer(dtype: TfdsDType) -> bool:
  return _is_dtype([np.integer], 'is_integer', dtype)


@py_utils.memoize()
def is_string(dtype: TfdsDType) -> bool:
  return _is_dtype([np.character, np.object_], np.object_, dtype)


@py_utils.memoize()
def is_same_dtype_type(a, b):
  if a == b:
    return True
  # NumPy strings can have different DTypes and yet be of the same DType type.
  return is_string(a) and is_string(b)


@py_utils.memoize()
def is_np_sub_dtype(value: np.dtype, super_type: np.dtype) -> bool:
  try:
    return np.issubdtype(value, super_type)
  except TypeError:
    return False



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VCTK speech synthesis dataset."""

import os

from absl import logging
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@misc{yamagishi2019vctk,
  author={Yamagishi, Junichi and Veaux, Christophe and MacDonald, Kirsten},
  title={{CSTR VCTK Corpus}: English Multi-speaker Corpus for {CSTR} Voice Cloning Toolkit (version 0.92)},
  publisher={University of Edinburgh. The Centre for Speech Technology Research (CSTR)},
  year=2019,
  doi={10.7488/ds/2645},
}
"""

_DESCRIPTION = """\
This CSTR VCTK Corpus includes speech data uttered by 110 English speakers with
various accents. Each speaker reads out about 400 sentences, which were selected
from a newspaper, the rainbow passage and an elicitation paragraph used for the
speech accent archive.

Note that the 'p315' text was lost due to a hard disk error.
"""

_URL = "https://doi.org/10.7488/ds/2645"
_DL_URL = "https://datashare.is.ed.ac.uk/bitstream/handle/10283/3443/VCTK-Corpus-0.92.zip"


class Vctk(tfds.core.GeneratorBasedBuilder):
  """VCTK speech synthesis dataset.

  The dataset is broken into two separate configs, each containing audio
  recorded using different microphones.  They are split this way for consistency
  with previous releases which only included a single microphone.
  """

  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(
          name="mic1",
          description="""
              Audio recorded using an omni-directional microphone (DPA 4035).
              Contains very low frequency noises.

              This is the same audio released in previous versions of VCTK:
              https://doi.org/10.7488/ds/1994
          """,
          version=tfds.core.Version("1.0.1"),
          release_notes={
              "1.0.0": "VCTK release 0.92.0.",
              "1.0.1": "Fix speech data type with dtype=tf.int16.",
          },
      ),
      tfds.core.BuilderConfig(
          name="mic2",
          description="""
              Audio recorded using a small diaphragm condenser microphone with
              very wide bandwidth (Sennheiser MKH 800).

              Two speakers, p280 and p315 had technical issues of the audio
              recordings using MKH 800.
          """,
          version=tfds.core.Version("1.0.1"),
          release_notes={
              "1.0.0": "VCTK release 0.92.0.",
              "1.0.1": "Fix speech data type with dtype=tf.int16.",
          },
      ),
  ]

  def _info(self):
    speaker_list = [
        "p225",
        "p226",
        "p227",
        "p228",
        "p229",
        "p230",
        "p231",
        "p232",
        "p233",
        "p234",
        "p236",
        "p237",
        "p238",
        "p239",
        "p240",
        "p241",
        "p243",
        "p244",
        "p245",
        "p246",
        "p247",
        "p248",
        "p249",
        "p250",
        "p251",
        "p252",
        "p253",
        "p254",
        "p255",
        "p256",
        "p257",
        "p258",
        "p259",
        "p260",
        "p261",
        "p262",
        "p263",
        "p264",
        "p265",
        "p266",
        "p267",
        "p268",
        "p269",
        "p270",
        "p271",
        "p272",
        "p273",
        "p274",
        "p275",
        "p276",
        "p277",
        "p278",
        "p279",
        "p280",
        "p281",
        "p282",
        "p283",
        "p284",
        "p285",
        "p286",
        "p287",
        "p288",
        "p292",
        "p293",
        "p294",
        "p295",
        "p297",
        "p298",
        "p299",
        "p300",
        "p301",
        "p302",
        "p303",
        "p304",
        "p305",
        "p306",
        "p307",
        "p308",
        "p310",
        "p311",
        "p312",
        "p313",
        "p314",
        "p315",
        "p316",
        "p317",
        "p318",
        "p323",
        "p326",
        "p329",
        "p330",
        "p333",
        "p334",
        "p335",
        "p336",
        "p339",
        "p340",
        "p341",
        "p343",
        "p345",
        "p347",
        "p351",
        "p360",
        "p361",
        "p362",
        "p363",
        "p364",
        "p374",
        "p376",
        "s5",
    ]
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "id": np.str_,
            "text": tfds.features.Text(),
            "speech": tfds.features.Audio(sample_rate=48000, dtype=np.int16),
            "speaker": tfds.features.ClassLabel(names=speaker_list),
            "gender": tfds.features.ClassLabel(names=["F", "M"]),
            "accent": tfds.features.ClassLabel(
                names=[
                    "American",
                    "Australian",
                    "British",
                    "Canadian",
                    "English",
                    "Indian",
                    "Irish",
                    "NewZealand",
                    "NorthernIrish",
                    "Scottish",
                    "SouthAfrican",
                    "Unknown",
                    "Welsh",
                ]
            ),
        }),
        supervised_keys=("text", "speech"),
        homepage=_URL,
        citation=_CITATION,
        metadata=tfds.core.MetadataDict(),
    )

  def _split_generators(self, dl_manager):
    extracted_dir = dl_manager.download_and_extract(_DL_URL)
    self._populate_metadata(extracted_dir)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"extracted_dir": extracted_dir},
        ),
    ]

  def _populate_metadata(self, extracted_dir):
    path = os.path.join(extracted_dir, "speaker-info.txt")
    speaker_info = tf.io.gfile.GFile(path).read()
    speaker_to_gender = {}
    speaker_to_accent = {}
    for line in speaker_info.splitlines()[1:]:
      fields = line.split()
      speaker, gender, accent = fields[0], fields[2], fields[3]
      speaker_to_gender[speaker] = gender
      speaker_to_accent[speaker] = accent
    self.info.metadata["speaker_info"] = speaker_info
    self.info.metadata["speaker_to_gender"] = speaker_to_gender
    self.info.metadata["speaker_to_accent"] = speaker_to_accent

  def _generate_examples(self, extracted_dir):
    """Yields examples."""
    speech_dir = "wav48_silence_trimmed"
    mic = "_%s" % self.builder_config.name
    speech_glob = os.path.join(extracted_dir, speech_dir, "*", "*%s.flac" % mic)
    for speech_path in tf.io.gfile.glob(speech_glob):
      text_path = speech_path.replace(speech_dir, "txt").split(mic)[0] + ".txt"
      key, _ = os.path.splitext(os.path.basename(text_path))
      if tf.io.gfile.exists(text_path):
        text = tf.io.gfile.GFile(text_path).read().strip()
      else:
        logging.info("No transcript found for utterance %s", key)
        text = ""
      speaker = key.split("_")[0]
      example = {
          "id": key,
          "text": text,
          "speech": speech_path,
          "speaker": speaker,
          "gender": self.info.metadata["speaker_to_gender"][speaker],
          "accent": self.info.metadata["speaker_to_accent"][speaker],
      }
      yield key, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Source-based register."""

import collections
import dataclasses
import datetime
import functools
import hashlib
import json
import tempfile
from typing import Any, Iterator, List, Optional, Type

from absl import logging

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.community import cache
from tensorflow_datasets.core.community import dataset_sources as dataset_sources_lib
from tensorflow_datasets.core.community import load
from tensorflow_datasets.core.community import register_base
from tensorflow_datasets.core.download import checksums
from tensorflow_datasets.core.utils import gcs_utils

# Datasets are installed as `import tfds_community.<ns>.<ds>.<hash>`
_IMPORT_MODULE_NAME = 'tfds_community'
_METADATA_FILENAME = 'installation.json'


@dataclasses.dataclass(frozen=True, eq=True)
class DatasetPackage:
  """Dataset metadata (before installation), of a single dataset package.

  Contains the information required to fetch the dataset package.

  Attributes:
    name: Dataset name
    source: Source that contains the source code (e.g. `github://...`)
  """

  name: naming.DatasetName
  source: dataset_sources_lib.DatasetSource
  # Ideally, we should also save the version so `tfds.load('ns:ds/1.0.0')`
  # fetch a specific version (e.g. at an older commit).

  @classmethod
  def from_json(cls, data: utils.Json) -> 'DatasetPackage':
    """Factory which creates the cls from json."""
    return cls(
        name=naming.DatasetName(namespace_name=data['name']),
        source=dataset_sources_lib.DatasetSource.from_json(data['source']),
    )

  def to_json(self) -> utils.Json:
    """Exports the cls as json."""
    return {
        'name': str(self.name),
        'source': self.source.to_json(),
    }


@dataclasses.dataclass(frozen=True, eq=True)
class _InstalledPackage:
  """Dataset metadata (after installation), of a single dataset package.

  Contains the local informations of the installed dataset package. This is
  specific to the user.

  Attributes:
    package: Source of the dataset package.
    instalation_date: Date of installation of the package
    hash: base64 checksum of the installed files
  """

  package: DatasetPackage
  instalation_date: datetime.datetime
  hash: str

  @property
  def module_name(self) -> str:
    """Module name to import this dataset."""
    name = self.package.name
    return f'{_IMPORT_MODULE_NAME}.{name.namespace}.{name.name}.{self.hash}.{name.name}'

  @property
  def installation_path(self) -> epath.Path:
    """Local path of the package."""
    name = self.package.name
    sub_dir = f'{_IMPORT_MODULE_NAME}/{name.namespace}/{name.name}/{self.hash}'
    return cache.module_path() / sub_dir

  @classmethod
  def from_json(cls, data: utils.Json) -> '_InstalledPackage':
    """Factory which creates the cls from json."""
    return cls(
        package=DatasetPackage.from_json(data['package']),
        instalation_date=datetime.datetime.fromisoformat(
            data['instalation_date']
        ),
        hash=data['hash'],
    )

  def to_json(self) -> utils.Json:
    """Exports the cls as json."""
    return {
        'package': self.package.to_json(),
        'instalation_date': self.instalation_date.isoformat(),
        'hash': self.hash,
    }


# TODO(tfds): py3.9 Should be `UserDict[naming.DatasetName, _DatasetPackage]`
class _PackageIndex(collections.UserDict):
  """Package index.

  Package index is a `Dict[DatasetName, _DatasetPackage]` loaded from cache.
  It has an additional `.refresh()` method to update the local cache by
  querying the remote index (stored in `gs://tfds-data`).

  On disk, the package index is a simple list of datasets with their
  associated source:

  ```jsonl
  {"name": "kaggle:ds0", "source": "github://..."}
  {"name": "kaggle:ds1", "source": "github://..."}
  {"name": "tensorflow_graphics:shapenet", "source": "github://..."}
  [...]
  ```
  """

  def __init__(self, path: epath.PathLike):
    """Contructor.

    Args:
      path: Remote location of the package index (file containing the list of
        dataset packages)
    """
    super().__init__()
    self._remote_path: epath.Path = epath.Path(path)
    self._cached_path: epath.Path = (
        cache.cache_path() / 'community-datasets-list.jsonl'
    )

    # Pre-load the index from the cache
    if self._cached_path.exists():
      self._refresh_from_content(self._cached_path.read_text())

  def _refresh_from_content(self, content: str) -> None:
    """Update the index from the given `jsonl` content."""
    dataset_packages = [
        DatasetPackage.from_json(json.loads(line))
        for line in content.splitlines()
        if line.strip()
    ]
    self.clear()
    self.update({src.name: src for src in dataset_packages})

  def refresh(self) -> None:
    """Update the cache."""
    # Should have a timer to avoid refreshing the cache immediatelly
    # (and a force=True option to ignore this)
    # e.g. with os.path.getmtime(cached_path) - time.gmtime()

    try:
      content = self._remote_path.read_text()
    except gcs_utils.gcs_unavailable_exceptions() as e:
      # Do not crash if GCS access not available, but instead silently reuse
      # the cache.
      logging.info(
          'Could not refresh the package index (GCS unavailable): %s', e
      )
      return

    # If read was sucessful, update the cache with the new dataset list
    self._cached_path.write_text(content)  # pytype: disable=attribute-error  # attribute-variable-annotations
    self._refresh_from_content(content)


class PackageRegister(register_base.BaseRegister):
  """Dataset register based on a list of remotely stored datasets definitions.

  Package register is similar to a dataset package manager. It contains a
  package index containing the list of all registered datasets with their
  associated location.
  When a specific dataset is requested, `PackageRegister` will download
  and cache the original source code locally.

  Usage:

  ```python
  register = PackageRegister(path='/path/to/datasets-source-list.jsonl')

  # List all registered datasets: ['kaggle:ds0', 'kaggle:ds1',...]
  register.list_builders()

  # Load a specific dataset
  builder = register.builder('tensorflow_graphics:shapenet')
  ```
  """

  def __init__(self, path: epath.PathLike):
    """Contructor.

    Args:
      path: Path to the register files containing the list of dataset sources,
        forwarded to `_PackageIndex`
    """
    self._path = path

  @functools.cached_property
  def _package_index(self) -> _PackageIndex:
    """`Dict[DatasetName, _DatasetPackage]` containg the community datasets."""
    # Use property to lazy-initialize the cache (and create the tmp dir) only
    # if it is used.
    return _PackageIndex(self._path)

  def list_builders(self) -> List[str]:
    """Returns the list of registered builders."""
    if not self._package_index:  # Package index not loaded nor cached
      self._package_index.refresh()  # Try updating the index
    return sorted(str(name) for name in self._package_index)  # pylint: disable=not-an-iterable

  def list_dataset_references(self) -> Iterator[naming.DatasetReference]:
    for name in self.list_builders():
      yield naming.DatasetReference.from_tfds_name(tfds_name=name)

  def builder_cls(
      self,
      name: naming.DatasetName,
  ) -> Type[dataset_builder.DatasetBuilder]:
    """Returns the builder class."""
    # Download the dataset generation code, or reuse the cache
    # TODO(tfds): Should add the option to request a specific code version
    installed_dataset = _download_or_reuse_cache(
        name=name,
        package_index=self._package_index,
    )

    # Load the dataset from the module
    return load.builder_cls_from_module(installed_dataset.module_name)

  def builder(
      self,
      name: naming.DatasetName,
      **builder_kwargs: Any,
  ) -> dataset_builder.DatasetBuilder:
    """Returns the dataset builder."""
    return self.builder_cls(name)(**builder_kwargs)  # pytype: disable=not-instantiable


def list_ds_packages_for_namespace(
    namespace: str,
    path: epath.Path,
) -> List[DatasetPackage]:
  """Returns the dataset names found in a specific directory.

  Directories that contain code should have the following structure:

  ```
  <path>/
      <dataset0>/
          <dataset0>.py
      <dataset1>/
          <dataset1>.py
      ...
  ```

  Additional files or folders which are not detected as datasets will be
  ignored (e.g. `__init__.py`).

  Args:
    namespace: Namespace of the datasets
    path: The directory path containing the datasets.

  Returns:
    ds_packages: The dataset packages found in the directory (sorted for
      determinism).

  Raises:
    FileNotFoundError: If the path cannot be reached.
  """
  if not path.exists():
    # Should be fault-tolerant in the future
    raise FileNotFoundError(f'Could not find datasets at {path}')

  all_packages = []
  for ds_path in path.iterdir():
    source = get_dataset_source(ds_path)
    if source:
      pkg = DatasetPackage(
          name=naming.DatasetName(namespace=namespace, name=ds_path.name),
          source=source,
      )
      all_packages.append(pkg)

  return all_packages


def get_dataset_source(
    ds_path: epath.Path,
) -> Optional[dataset_sources_lib.DatasetSource]:
  """Returns a `DatasetSource` instance if the given path corresponds to a dataset.

  To determine whether the given path contains a dataset, a simple heuristic is
  used that checks whether the path has the following structure:

  ```
  <ds_name>/
      <ds_name>.py
  ```

  If so, all `.py`, `.txt`, `.tsv`, `.json` files will be added to the package.

  Args:
    ds_path: Path of the dataset module

  Returns:
    A `DatasetSource` instance if the path matches the expected file structure.
  """
  filter_list = {'__init__.py'}
  suffixes_list = ('.txt', '.tsv', '.py', '.json')

  def is_interesting_file(fname: str) -> bool:
    return fname.endswith(suffixes_list) and fname not in filter_list

  if not ds_path.is_dir():
    return None
  all_filenames = set(f.name for f in ds_path.iterdir())
  if f'{ds_path.name}.py' not in all_filenames:
    return None

  return dataset_sources_lib.DatasetSource(
      root_path=ds_path,
      filenames=sorted(
          [fname for fname in all_filenames if is_interesting_file(fname)]
      ),
  )


def _download_or_reuse_cache(
    name: naming.DatasetName,
    package_index: _PackageIndex,
) -> _InstalledPackage:
  """Downloads the dataset generation source code.

  Search the dataset in the cache, or download it from the package index
  otherwise.

  Args:
    name: Dataset name to load.
    package_index: Index of all community datasets. Might be updated.

  Returns:
    The installed dataset information.

  Raises:
    DatasetNotFoundError: If the dataset can't be loaded.
  """
  # Dataset can be:
  # * Installed locally (in the cache) -> reuse
  # * Not installed but present in the package index -> install
  # * Not present in the package index -> raise error

  # Check if the file is already downloaded/cached
  # TODO(tfds): To force a download even if file already present, we
  # should add a `ignore_cache=True` option in `tfds.load`. Or should always
  # try to download the file ?
  last_installed_version = _get_last_installed_version(name)
  if last_installed_version:
    return last_installed_version

  # If file isn't cached yet, we need to download it.
  # First need to find it's location.
  if name not in package_index:
    # If not, we need to update the package index cache
    package_index.refresh()
  # If the dataset is present in the package index cache, use this
  package = package_index.get(name)
  if not package:
    # If still not found, raise an DatasetNotFoundError
    raise registered.DatasetNotFoundError(
        f'Could not find dataset {name}: Dataset not found among the '
        f'{len(package_index)} datasets of the community index.'
    )

  # If package was found, download it.
  installed_package = _download_and_cache(package)
  return installed_package


def _get_last_installed_version(
    name: naming.DatasetName,
) -> Optional[_InstalledPackage]:
  """Checks whether the datasets is installed locally and returns it."""
  root_dir = (
      cache.module_path() / _IMPORT_MODULE_NAME / name.namespace / name.name
  )
  if not root_dir.exists():  # Dataset not found
    return None

  all_installed_package_metadatas = [
      package / _METADATA_FILENAME for package in root_dir.iterdir()
  ]
  all_installed_packages = [
      _InstalledPackage.from_json(json.loads(metadata.read_text()))
      for metadata in all_installed_package_metadatas
      if metadata.exists()
  ]
  all_installed_packages = sorted(
      all_installed_packages, key=lambda p: p.instalation_date
  )

  if not all_installed_packages:  # No valid package found
    return None
  else:
    return all_installed_packages[-1]  # Most recently installed package


def _download_and_cache(package: DatasetPackage) -> _InstalledPackage:
  """Downloads and installs locally the dataset source.

  This function install the dataset package in:
  `<module_path>/<namespace>/<ds_name>/<hash>/...`.

  Args:
    package: Package to install.

  Returns:
    installed_dataset: The installed dataset package.
  """
  tmp_dir = epath.Path(tempfile.mkdtemp())
  try:
    # Download the package in a tmp directory
    dataset_sources_lib.download_from_source(
        package.source,
        tmp_dir,
    )

    # Compute the package hash (to install the dataset in a unique dir)
    package_hash = _compute_dir_hash(tmp_dir)

    # Add package metadata
    installed_package = _InstalledPackage(
        package=package,
        instalation_date=datetime.datetime.now(),
        hash=package_hash,
    )
    package_metadata = json.dumps(installed_package.to_json())
    (tmp_dir / _METADATA_FILENAME).write_text(package_metadata)

    # Rename the package to it's final destination
    installation_path = installed_package.installation_path
    if installation_path.exists():  # Package already exists (with same hash)
      # In the future, we should be smarter to allow overwrite.
      raise ValueError(
          f'Package {package} already installed in {installation_path}.'
      )
    installation_path.parent.mkdir(parents=True, exist_ok=True)
    tmp_dir.rename(installation_path)
  finally:
    # Cleanup the tmp directory if it still exists.
    if tmp_dir.exists():
      tmp_dir.rmtree()

  return installed_package


def _compute_dir_hash(path: epath.Path) -> str:
  """Computes the checksums of the given directory deterministically."""
  all_files = sorted(path.iterdir())

  if any(f.is_dir() for f in all_files):
    raise ValueError('Installed package should only contains files.')

  # Concatenate the filenames and files content to create the directory hash
  all_checksums = [f.name for f in all_files]
  all_checksums += [checksums.compute_url_info(f).checksum for f in all_files]
  return hashlib.sha256(''.join(all_checksums).encode()).hexdigest()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy import utils.
"""

from __future__ import annotations

import builtins
import contextlib
import dataclasses
import functools
import importlib
import time
import types
from typing import Any, Callable, Iterator, Optional, Tuple

Callback = Callable[..., None]


@dataclasses.dataclass
class LazyModule:
  """Module loaded lazily during first call."""

  module_name: str
  module: Optional[types.ModuleType] = None
  fromlist: Optional[Tuple[str, ...]] = ()
  error_callback: Optional[Callback] = None
  success_callback: Optional[Callback] = None

  @classmethod
  @functools.lru_cache(maxsize=None)
  def from_cache(cls, **kwargs):
    """Factory to cache all instances of module.

    Note: The cache is global to all instances of the
    `lazy_imports` context manager.

    Args:
      **kwargs: Init kwargs

    Returns:
      New object
    """
    return cls(**kwargs)

  def __getattr__(self, name: str) -> Any:
    if name in self.fromlist:
      module_name = f"{self.module_name}.{name}"
      return self.from_cache(
          module_name=module_name,
          module=self.module,
          fromlist=self.fromlist,
          error_callback=self.error_callback,
          success_callback=self.success_callback,
      )
    if self.module is None:  # Load on first call
      try:
        start_import_time = time.perf_counter()
        self.module = importlib.import_module(self.module_name)
        import_time_ms = (time.perf_counter() - start_import_time) * 1000
        if self.success_callback is not None:
          self.success_callback(
              import_time_ms=import_time_ms,
              module=self.module,
              module_name=self.module_name,
          )
      except ImportError as exception:
        if self.error_callback is not None:
          self.error_callback(exception=exception, module_name=self.module_name)
        raise exception
    return getattr(self.module, name)


@contextlib.contextmanager
def lazy_imports(
    error_callback: Optional[Callback] = None,
    success_callback: Optional[Callback] = None,
) -> Iterator[None]:
  """Context Manager which lazy loads packages.

  Their import is not executed immediately, but is postponed to the first
  call of one of their attributes.

  Warning: mind current implementation's limitations:

  - You can only lazy load modules (`from x import y` will not work if `y` is a
    constant or a function or a class).
  - You cannot `import x.y` if `y` is not imported in the `x/__init__.py`.

  Usage:

  ```python
  from tensorflow_datasets.core.utils.lazy_imports_utils import lazy_imports

  with lazy_imports():
    import tensorflow as tf
  ```

  Args:
    error_callback: a callback to trigger when an import fails. The callback is
      passed kwargs containing: 1) exception (ImportError): the exception that
      was raised after the error; 2) module_name (str): the name of the imported
      module.
    success_callback: a callback to trigger when an import succeeds. The
      callback is passed kwargs containing: 1) import_time_ms (float): the
      import time (in milliseconds); 2) module (Any): the imported module; 3)
      module_name (str): the name of the imported module.

  Yields:
    None
  """
  # Need to mock `__import__` (instead of `sys.meta_path`, as we do not want
  # to modify the `sys.modules` cache in any way)
  original_import = builtins.__import__
  try:
    builtins.__import__ = functools.partial(
        _lazy_import,
        error_callback=error_callback,
        success_callback=success_callback,
    )
    yield
  finally:
    builtins.__import__ = original_import


def _lazy_import(
    name: str,
    globals_=None,
    locals_=None,
    fromlist: tuple[str, ...] = (),
    level: int = 0,
    *,
    error_callback: Optional[Callback],
    success_callback: Optional[Callback],
):
  """Mock of `builtins.__import__`."""
  del globals_, locals_  # Unused

  if level:
    raise ValueError(f"Relative import statements not supported ({name}).")

  if not fromlist:
    # import x.y.z
    # import x.y.z as z
    # In that case, Python would usually import the entirety of `x` if each
    # submodule is imported in its parent's `__init__.py`. So we do the same.
    root_name = name.split(".")[0]
    return LazyModule.from_cache(
        module_name=root_name,
        error_callback=error_callback,
        success_callback=success_callback,
    )
  # from x.y.z import a, b
  return LazyModule.from_cache(
      module_name=name,
      fromlist=fromlist,
      error_callback=error_callback,
      success_callback=success_callback,
  )


def tf_error_callback(**kwargs):
  del kwargs
  print("\n\n***************************************************************")
  print(
      "Failed to import TensorFlow. Please note that TensorFlow is not "
      "installed by default when you install TFDS. This allows you "
      "to choose to install either `tf-nightly` or `tensorflow`. "
      "Please install the most recent version of TensorFlow, by "
      "following instructions at https://tensorflow.org/install."
  )
  print("***************************************************************\n\n")


MIN_TF_VERSION = "2.1.0"

_ensure_tf_version_called = False


# Ensure TensorFlow version is sufficiently recent. This needs to happen after
# TensorFlow was imported, as it is referenced.
def ensure_tf_version(**kwargs):
  """Ensures TensorFlow version is sufficient."""
  # Only check the first time.
  global _ensure_tf_version_called
  if _ensure_tf_version_called:
    return
  _ensure_tf_version_called = True

  tf_version = kwargs["module"].__version__
  if tf_version < MIN_TF_VERSION:
    raise ImportError(
        "This version of TensorFlow Datasets requires TensorFlow "
        f"version >= {MIN_TF_VERSION}; Detected an installation of version "
        f"{tf_version}. Please upgrade TensorFlow to proceed."
    )


def array_record_error_callback(**kwargs):
  del kwargs
  print("\n\n***************************************************************")
  print(
      "Failed to import ArrayRecord. This probably means that you are running"
      " on macOS or Windows. ArrayRecord currently does not work for your"
      " infrastructure, because it uses Python bindings in C++. We are actively"
      " working on this issue. Thanks for your understanding."
  )
  print("***************************************************************\n\n")


with lazy_imports():
  import pandas  # pylint: disable=g-import-not-at-top,unused-import


with lazy_imports(
    error_callback=tf_error_callback, success_callback=ensure_tf_version
):
  import tensorflow  # pylint: disable=g-import-not-at-top,unused-import  # pytype: disable=import-error


with lazy_imports(error_callback=array_record_error_callback):
  from array_record.python import array_record_data_source  # pylint: disable=g-import-not-at-top,unused-import
  from array_record.python import array_record_module  # pylint: disable=g-import-not-at-top,unused-import



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for WMT translate dataset module."""

from tensorflow_datasets import testing
from tensorflow_datasets.translate import wmt19


class TranslateDeEnWmt19Test(testing.DatasetBuilderTestCase):
  DATASET_CLASS = wmt19.Wmt19Translate
  BUILDER_CONFIG_NAMES_TO_TEST = ["de-en"]
  OVERLAPPING_SPLITS = ["validation"]

  DL_EXTRACT_RESULT = {
      "paracrawl_v3": ["sentences.de-en.tmx"],
      "newscommentary_v14": ["sentences.de-en.tsv"],
      "wikititles_v1": ["sentences.de-en.tsv"],
      "rapid_2019": ["rapid_2019"],
      "newstest2018": ["validation"],
  }

  SEQUENTUAL_DL_EXTRACT_RESULT = {
      "http://www.statmt.org/wmt13/training-parallel-commoncrawl.tgz": (
          "commoncrawl"
      ),
      "http://www.statmt.org/europarl/v9/training/europarl-v9.de-en.tsv.gz": (
          "sentences.de-en.tsv"
      ),
  }

  def _get_dl_extract_result(self, url):
    if not url:
      return []
    if isinstance(url, dict):
      return super()._get_dl_extract_result(url)
    return self.dummy_data / self.SEQUENTUAL_DL_EXTRACT_RESULT[url]

  SPLITS = {
      "train": 12,
      "validation": 2,
  }


class TranslateCsEnWmt19Test(testing.DatasetBuilderTestCase):
  DATASET_CLASS = wmt19.Wmt19Translate
  BUILDER_CONFIG_NAMES_TO_TEST = ["cs-en"]
  OVERLAPPING_SPLITS = ["validation"]

  DL_EXTRACT_RESULT = {
      "czeng17_filter": ["czeng"],
      "paracrawl_v3": ["sentences.cs-en.tmx"],
      "newscommentary_v14": ["sentences.cs-en.tsv"],
      "wikititles_v1": ["sentences.cs-en.tsv"],
      "newstest2018": ["validation"],
  }

  SEQUENTUAL_DL_EXTRACT_RESULT = {
      "http://www.statmt.org/wmt13/training-parallel-commoncrawl.tgz": (
          "commoncrawl"
      ),
      "http://www.statmt.org/europarl/v9/training/europarl-v9.cs-en.tsv.gz": (
          "sentences.cs-en.tsv"
      ),
  }

  def _get_dl_extract_result(self, url):
    if not url:
      return []
    if isinstance(url, dict):
      return super()._get_dl_extract_result(url)
    return self.dummy_data / self.SEQUENTUAL_DL_EXTRACT_RESULT[url]

  SPLITS = {
      "train": 13,
      "validation": 2,
  }


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Geirhos conflict stimulus set."""

import os
import re

from etils import epath
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@inproceedings{
  geirhos2018imagenettrained,
  title={ImageNet-trained {CNN}s are biased towards texture; increasing shape
         bias improves accuracy and robustness.},
  author={Robert Geirhos and Patricia Rubisch and Claudio Michaelis and
          Matthias Bethge and Felix A. Wichmann and Wieland Brendel},
  booktitle={International Conference on Learning Representations},
  year={2019},
  url={https://openreview.net/forum?id=Bygh9j09KX},
}
"""

_DESCRIPTION = """\
Shape/texture conflict stimuli from "ImageNet-trained CNNs are biased towards \
texture; increasing shape bias improves accuracy and robustness."

Note that, although the dataset source contains images with matching shape and
texture and we include them here, they are ignored for most evaluations in the
original paper.
"""

_BASE_URL = "https://github.com/rgeirhos/texture-vs-shape"
_DOWNLOAD_URL = "https://github.com/rgeirhos/texture-vs-shape/archive/1b69c6a445c3348927139edb30a5134521fd4b03.zip"
_IMAGENET_MAPPING_URL = "https://raw.githubusercontent.com/rgeirhos/generalisation-humans-DNNs/5bbe08f6821e1eb2bdbe98acebf3586d36be00ab/16-class-ImageNet/MSCOCO_to_ImageNet_category_mapping.txt"
_DATA_DIR_PATH = "texture-vs-shape-1b69c6a445c3348927139edb30a5134521fd4b03/stimuli/style-transfer-preprocessed-512"

_CLASSES = [
    "airplane",
    "bear",
    "bicycle",
    "bird",
    "boat",
    "bottle",
    "car",
    "cat",
    "chair",
    "clock",
    "dog",
    "elephant",
    "keyboard",
    "knife",
    "oven",
    "truck",
]


class GeirhosConflictStimuli(tfds.core.GeneratorBasedBuilder):
  """Shape/Texture conflict ."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    """Define dataset info."""

    imagenet_names_file = imagenet_common.label_names_file()
    return tfds.core.DatasetInfo(
        builder=self,
        description=(_DESCRIPTION),
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(),
            "shape_label": tfds.features.ClassLabel(names=_CLASSES),
            "shape_imagenet_labels": tfds.features.Sequence(
                tfds.features.ClassLabel(names_file=imagenet_names_file)
            ),
            "texture_label": tfds.features.ClassLabel(names=_CLASSES),
            "texture_imagenet_labels": tfds.features.Sequence(
                tfds.features.ClassLabel(names_file=imagenet_names_file)
            ),
            "file_name": tfds.features.Text(),
        }),
        supervised_keys=("image", "shape_label"),
        homepage=_BASE_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Define splits."""

    dl_paths = dl_manager.download_and_extract({
        "texture_vs_shape": _DOWNLOAD_URL,
        "imagenet_mapping": _IMAGENET_MAPPING_URL,
    })

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "data_dir_path": os.path.join(
                    dl_paths["texture_vs_shape"], _DATA_DIR_PATH
                ),
                "imagenet_mapping_path": dl_paths["imagenet_mapping"],
            },
        ),
    ]

  def _generate_examples(self, data_dir_path, imagenet_mapping_path):
    """Generate images and labels for splits."""
    # Read ImageNet mapping.
    imagenet_names = set(self.info.features["shape_imagenet_labels"].names)

    with epath.Path(imagenet_mapping_path).open() as f:
      mapping_txt = f.read()
    mapping = {}
    for match in re.finditer(r"([a-z]+)\s*=\s*\[([^\]]+)\]", mapping_txt):
      mapping[match.group(1)] = list(
          sorted(
              imagenet_names.intersection(
                  re.sub(r"\s", "", match.group(2)).split(",")
              )
          )
      )

    # Process images.
    for shape_class_name in tf.io.gfile.listdir(data_dir_path):
      class_dir_path = os.path.join(data_dir_path, shape_class_name)
      for image_name in tf.io.gfile.listdir(class_dir_path):
        image = os.path.join(class_dir_path, image_name)
        texture_class_name = re.search("-([a-z]+)", image_name).group(1)  # pytype: disable=attribute-error
        yield image_name, {
            "image": image,
            "file_name": image_name,
            "shape_label": shape_class_name,
            "shape_imagenet_labels": mapping[shape_class_name],
            "texture_label": texture_class_name,
            "texture_imagenet_labels": mapping[texture_class_name],
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for subsplits_utils."""

import pytest
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import subsplits_utils


def _filename_template(split: str) -> naming.ShardedFileTemplate:
  return naming.ShardedFileTemplate(
      dataset_name='mnist',
      split=split,
      filetype_suffix='tfrecord',
      data_dir='/path',
  )


@pytest.mark.parametrize(
    'num_examples, n, drop_remainder, expected',
    [
        (9, 1, True, ['']),  # Full split selected
        (9, 2, True, ['[:4]', '[4:8]']),
        (9, 3, True, ['[:3]', '[3:6]', '[6:9]']),
        (9, 4, True, ['[:2]', '[2:4]', '[4:6]', '[6:8]']),  # Last ex dropped
        (11, 2, True, ['[:5]', '[5:10]']),
        (11, 3, True, ['[:3]', '[3:6]', '[6:9]']),  # Last 2 exs dropped
        (9, 1, False, ['']),
        (9, 2, False, ['[:5]', '[5:9]']),  # split0 has extra ex
        (9, 3, False, ['[:3]', '[3:6]', '[6:9]']),
        (9, 4, False, ['[:3]', '[3:5]', '[5:7]', '[7:9]']),  # 0 has extra ex
        (11, 3, False, ['[:4]', '[4:8]', '[8:11]']),  # 0, 1 have extra ex
        (11, 4, False, ['[:3]', '[3:6]', '[6:9]', '[9:11]']),
    ],
)
def test_even_splits(num_examples, n, drop_remainder, expected):
  split_infos = splits_lib.SplitDict(
      split_infos=[
          splits_lib.SplitInfo(
              name='train',
              shard_lengths=[num_examples],
              num_bytes=0,
              filename_template=_filename_template('train'),
          ),
      ]
  )

  subsplits = subsplits_utils.even_splits(
      'train', n, drop_remainder=drop_remainder
  )

  file_instructions = [split_infos[s].file_instructions for s in subsplits]
  expected_file_instructions = [
      split_infos[f'train{s}'].file_instructions for s in expected
  ]
  assert file_instructions == expected_file_instructions


def test_even_splits_subsplit():
  split_infos = splits_lib.SplitDict(
      split_infos=[
          splits_lib.SplitInfo(
              name='train',
              shard_lengths=[2, 3, 2, 3],  # 10
              num_bytes=0,
              filename_template=_filename_template('train'),
          ),
          splits_lib.SplitInfo(
              name='test',
              shard_lengths=[8],
              num_bytes=0,
              filename_template=_filename_template('test'),
          ),
      ]
  )

  # Test to split multiple splits
  subsplits = subsplits_utils.even_splits('train+test[50%:]', 3)

  expected = [
      'train[:4]+test[4:6]',
      'train[4:7]+test[6:7]',
      'train[7:]+test[7:8]',
  ]

  file_instructions = [split_infos[s].file_instructions for s in subsplits]
  expected_file_instructions = [
      split_infos[s].file_instructions for s in expected
  ]
  assert file_instructions == expected_file_instructions


def test_even_splits_add():
  # Compatibility of even_splits with other splits

  split_infos = splits_lib.SplitDict(
      split_infos=[
          splits_lib.SplitInfo(
              name='train',
              shard_lengths=[2, 3, 2, 3],  # 10
              num_bytes=0,
              filename_template=_filename_template('train'),
          ),
          splits_lib.SplitInfo(
              name='test',
              shard_lengths=[8],
              num_bytes=0,
              filename_template=_filename_template('test'),
          ),
          splits_lib.SplitInfo(
              name='validation',
              shard_lengths=[8],
              filename_template=_filename_template('validation'),
              num_bytes=0,
          ),
      ]
  )

  # Test to split multiple splits
  split = subsplits_utils.even_splits('train', 3, drop_remainder=True)[0]
  split = split + 'test'

  expected = 'train[:3]+test'

  file_instructions = split_infos[split].file_instructions
  expected_file_instructions = split_infos[expected].file_instructions
  assert file_instructions == expected_file_instructions

  # Test nested `even_splits`
  splits = subsplits_utils.even_splits('validation', n=2)
  splits = subsplits_utils.even_splits(splits[1], n=2)
  assert (
      split_infos[splits[0]].file_instructions
      == split_infos['validation[4:6]'].file_instructions
  )
  assert (
      split_infos[splits[1]].file_instructions
      == split_infos['validation[6:8]'].file_instructions
  )


def test_split_for_jax_process():
  split = subsplits_utils.split_for_jax_process('train')
  assert isinstance(split, subsplits_utils._EvenSplit)
  assert split.split == 'train'
  assert split.index == 0
  assert split.count == 1



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.video_feature."""

import json
import os
import pathlib

from etils import epath
import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features


class VideoFeatureTest(testing.FeatureExpectationsTestCase):

  @property
  def _test_data_path(self):
    return os.path.join(os.path.dirname(__file__), '../../testing/test_data')

  def test_video_numpy(self):
    np_video = np.random.randint(256, size=(128, 64, 64, 3), dtype=np.uint8)

    self.assertFeature(
        feature=features.Video(shape=(None, 64, 64, 3)),
        shape=(None, 64, 64, 3),
        dtype=tf.uint8,
        tests=[
            testing.FeatureExpectationItem(
                value=np_video,
                expected=np_video,
            ),
        ],
        test_attributes=dict(_encoding_format='png', _extra_ffmpeg_args=[]),
    )

  def test_video_concatenated_frames(self):
    video_shape = (None, 400, 640, 3)
    lsun_examples_path = os.path.join(self._test_data_path, 'lsun_examples')
    frames_paths = [
        os.path.join(lsun_examples_path, '{}.jpg'.format(i))
        for i in (1, 2, 3, 4)
    ]
    frames = []
    for frame_path in frames_paths:
      with tf.io.gfile.GFile(frame_path, 'rb') as frame_fp:
        frames.append(tf.image.decode_jpeg(frame_fp.read(), channels=3))
    video = tf.stack(frames)

    # Check both str and Path objects
    frames_paths[-1] = pathlib.Path(frames_paths[-1])

    self.assertFeature(
        feature=features.Video(shape=video_shape),
        shape=video_shape,
        dtype=tf.uint8,
        tests=[
            # Numpy array
            testing.FeatureExpectationItem(
                value=frames_paths,
                expected=video,
            ),
        ],
    )

  def test_video_ffmpeg(self):
    video_path = os.path.join(self._test_data_path, 'video.mkv')
    video_json_path = os.path.join(self._test_data_path, 'video.json')
    with epath.Path(video_json_path).open() as fp:
      video_array = np.asarray(json.load(fp))

    self.assertFeature(
        feature=features.Video(shape=(5, 4, 2, 3)),
        shape=(5, 4, 2, 3),
        dtype=tf.uint8,
        tests=[
            testing.FeatureExpectationItem(
                value=video_path,
                expected=video_array,
            ),
            testing.FeatureExpectationItem(
                value=pathlib.Path(video_path),
                expected=video_array,
            ),
        ],
    )

    class GFileWithSeekOnRead(tf.io.gfile.GFile):

      def read(self, *args, **kwargs):
        data_read = super(GFileWithSeekOnRead, self).read(*args, **kwargs)
        self.seek(0)
        return data_read

    with GFileWithSeekOnRead(video_path, 'rb') as video_fp:
      self.assertFeature(
          feature=features.Video(shape=(5, 4, 2, 3)),
          shape=(5, 4, 2, 3),
          dtype=tf.uint8,
          tests=[
              testing.FeatureExpectationItem(
                  value=video_fp,
                  expected=video_array,
              ),
          ],
      )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_datasets/proto/tf_example.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow_datasets.proto import tf_feature_generated_pb2 as tensorflow__datasets_dot_proto_dot_tf__feature__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*tensorflow_datasets/proto/tf_example.proto\x12\x0ftensorflow_copy\x1a*tensorflow_datasets/proto/tf_feature.proto"6\n\x07\x45xample\x12+\n\x08\x66\x65\x61tures\x18\x01'
    b' \x01(\x0b\x32\x19.tensorflow_copy.Features"s\n\x0fSequenceExample\x12*\n\x07\x63ontext\x18\x01'
    b' \x01(\x0b\x32\x19.tensorflow_copy.Features\x12\x34\n\rfeature_lists\x18\x02'
    b' \x01(\x0b\x32\x1d.tensorflow_copy.FeatureListsB\x81\x01\n\x16org.tensorflow.exampleB\rExampleProtosP\x01ZSgithub.com/tensorflow/tensorflow/tensorflow/go/core/example/example_protos_go_proto\xf8\x01\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'tensorflow_datasets.proto.tf_example_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026org.tensorflow.exampleB\rExampleProtosP\001ZSgithub.com/tensorflow/tensorflow/tensorflow/go/core/example/example_protos_go_proto\370\001\001'
  _EXAMPLE._serialized_start = 107
  _EXAMPLE._serialized_end = 161
  _SEQUENCEEXAMPLE._serialized_start = 163
  _SEQUENCEEXAMPLE._serialized_end = 278
# @@protoc_insertion_point(module_scope)
